"""Tutorial to use the GroundTruth class of fmdt.truth.GroundTruth"""

import fmdt
import fmdt.truth

# We need to define the directory where our GroundTruth videos are found.

vid_dir12 = "/home/ejovo/Videos/Watec12mm"
vid_dir6  = "/home/ejovo/Videos/Watec6mm"

csv12 = "./human_detections_draco12.csv"
csv6  = "./human_detections_draco6.csv"

# We can create a new `GroundTruth` object - a list of `HumanDetection` paired
# with useful functions - two ways

gt12 = fmdt.truth.GroundTruth(csv=csv12, vid_dir=vid_dir12)
gt6  = fmdt.truth.GroundTruth(csv=csv6,  vid_dir=vid_dir6)

print(type(gt12))
print(gt12)

print(type(gt6))
print(gt6)

# We can create an Args and run that command across all the 
# videos in our database with GroundTruth.try_command

d_args = {
    "light_min": 251,
    "light_max": 252,
    "trk_out_path": "tracks.txt",
    "timeout": 1, #! Special parameter of the pyhton SUBPROCESS managing fmdt-detect
}

args = fmdt.Args(detect_args=d_args)

success = gt6.try_command(args)

print(success)
print(sum(success))


import unittest
import shutil
# import fmdt.api
import os
import fmdt.args
import fmdt.api
import fmdt.res
import fmdt.core
import fmdt.download
import pandas as pd

class TestAPI(unittest.TestCase):

    def test_fmdt_detect_existence(self):
        fmdt_detect_exe = shutil.which("fmdt-detect")
        self.assertTrue(not fmdt_detect_exe is None, "Executable 'fmdt-detect' not found on the path")

    
class TestARGS(unittest.TestCase):

    #================= Creation of Args type ==============
    def test_detect_arg_creation(self):

        args = fmdt.detect_args("demo.mp4")

        self.assertTrue(args.detect_args.vid_in_path == "demo.mp4")
        self.assertTrue(args.visu_args.vid_out_path == "demo_visu.mp4")

    def test_arg_creation(self):

        args = fmdt.Args.new(vid_in_path="demo.mp4", vid_out_path="demo_out.mp4", vid_in_start=10)

        self.assertTrue(args.visu_args.vid_out_path == "demo_out.mp4")
        self.assertTrue(args.visu_args.vid_in_start == 10)
        self.assertTrue(args.detect_args.vid_in_start == 10)

        args2 = fmdt.Args.new(vid_in_path="demo.mp4", timeout=4, log=True)

        self.assertTrue(args2.timeout == 4)
        self.assertTrue(args2.log)

class TestDetections(unittest.TestCase):

    def test_demo_plain(self):

        if not os.path.exists("demo.mp4"):
            print("Cannot run detection tests on demo.mp4 because it was not found")
            exit(2)
        res = fmdt.detect("demo.mp4")

        """Needs to be run with "demo.mp4" in the current working directory"""

        self.assertTrue(isinstance(res, fmdt.res.DetectionResult))
        self.assertTrue(isinstance(res.trk_list[0], fmdt.core.TrackedObject))

        self.assertTrue(len(res.trk_list) == 38)
        self.assertTrue(res.trk_list[0].type_str() == "Meteor")
        self.assertTrue(res.nframes == 256)

        # Since we haven't passed the log_path parameter, we expect no data to be stored
        self.assertTrue(len(res.nrois) == 0)

    def test_demo_log(self):

        if not os.path.exists("demo.mp4"):
            print("Cannot run detection tests on demo.mp4 because it was not found")
            exit(2)
        res = fmdt.detect("demo.mp4", log_path="log")

        self.assertTrue(len(res.nrois) == 256)

        df = res.to_dataframe()

        self.assertTrue(len(df) == 256)

    def test_demo_log_all(self):

        if not os.path.exists("demo.mp4"):
            print("Cannot run detection tests on demo.mp4 because it was not found")
            exit(2)
        res = fmdt.detect("demo.mp4", log_path="log", trk_all=True)


        self.assertTrue(len(res.trk_list) == 82)
        self.assertTrue(len(res.nrois) == 256)
        

        df = res.to_dataframe()

        self.assertTrue(len(df) == 256)


class TestGroundTruth(unittest.TestCase):

    def test_downloads(self):
        """Download all the csvs associated with our data base"""

        fmdt.download.download_csvs(overwrite=True)

        # Now I should be able to load in the values using pandas I guess
        # We want to load in some ground truth values as well
        gt6 = fmdt.GroundTruth("human_detections_draco6.csv")
        gt12 = fmdt.GroundTruth("human_detections_draco12.csv")
        gtall = fmdt.GroundTruth("human_detections.csv")

        self.assertTrue(len(gt6) == 38)
        self.assertTrue(len(gt12) == 39)
        self.assertTrue(len(gtall) == 77)

# class TestKnownResults(unittest.TestCase):

#     def test_


def main() -> None:
    unittest.main()    

if __name__ == '__main__':
    main()


import fmdt
import fmdt.truth

vid = "Draconids-6mm1.00-2750-163200.avi"

light_min = 186
light_max = 211

human = fmdt.truth.HumanDetection(vid, 45, 59, 240, 226, 125, 264)

track = "tracks.txt"
bb    = "bb.txt"

# res = fmdt.detect(vid, light_min=light_min, light_max=light_max, trk_all=True, trk_out_path=track, trk_bb_path=bb)

# fmdt.visu(vid, track, bb, "visu.mp4", True)

# Get the tracking list
print(fmdt.extract_all_information(track))

tracking_list = fmdt.extract_all_information(track)

for t in tracking_list:
    print(t)

is_det = fmdt.truth.is_meteor_detected(human, tracking_list)
print(f"Ground truth: {human} detected in {tracking_list}? : {is_det}")


import os
import fmdt
import sys

# Directory whose videos you would like to view
# directory = "/run/media/ejovo/Seagate Portable Drive/Meteors/Watec12mm/Meteor/"
directory = "/home/ejovo/Videos/Watec12mm"

def print_help() -> None:
    print("Type the name of the directory you wish to process:")
    # print(f"Ex: {sys.argv[0]} ~/Videos/Meteors")

if directory is None and len(sys.argv) == 1:
    print_help()
    directory = input()
    # assert False
elif directory is None:
    directory = sys.argv[1]

args = {
    "trk_out_path": "tracks.txt",
    "light_min": 190,
    "light_max": 235, # None of the videos fail with _just_ a light max
    "trk_all": True,
    "log": True
}

a = fmdt.Args(detect_args=args)

# List the files in the directory

print(os.listdir(directory))

print("Running fmdt.detect_directory with args:")
print(a.detect_args)

fmdt.detect_directory(directory, a, True)


"""Module dedicated to the management of config files and stuff"""
import jsonpickle
import os
from appdirs import AppDirs
import fmdt.truth
import fmdt.download
import fmdt.res
import pandas as pd
import shutil

dirs = AppDirs("fmdt_python")
config_file = "config.json"
full_path = dirs.user_data_dir + "/" + config_file

if not os.path.exists(dirs.user_data_dir):
    os.mkdir(dirs.user_data_dir)

class Config:

    def __init__(self, draco6_dir, draco12_dir, windows_dir):
        self.d6 = draco6_dir
        self.d12 = draco12_dir
        self.win = windows_dir

    def __str__(self) -> str:
        s0= "============================\n"
        h = "           Config           \n"
        s0= "============================\n"
        s = f"Draco6   {self.d6}\nDraco12  {self.d12}\nWindow   {self.win}"
        return h + s0 + s
    
    def __repr__(self) -> str:
        return self.__str__()

    def to_json(self) -> str:
        return jsonpickle.encode(self)

    def save(self):

        with open(full_path, "w") as file:
            file.write(self.to_json())
        
        print(f"Saved config to {full_path}")

    @staticmethod
    def from_json(json: str):
        return jsonpickle.decode(json)

    @staticmethod
    def load():
        return load_config()
    
    @staticmethod
    def gt6_csv() -> str:
        return dirs.user_data_dir + "/gt6.csv"
    
    @staticmethod
    def gt12_csv() -> str:
        return dirs.user_data_dir + "/gt12.csv"
    
    # @staticmethod
    # def gt612_csv() -> str:
    #     return dirs.user_data_dir + "/gt12.csv"

def check_for_config_file() -> bool:
    """Return True is a fmdt_config.json file exists in the user_data_dir"""
    return os.path.exists(full_path)

def load_config() -> Config:
    if check_for_config_file():
        with open(full_path, "r") as file:
            content = file.read()
            return Config.from_json(content)
    else: # WE HAVENT SET UP OUR CONFIG!!!

        raise Exception(f"""
            ============================================================
                            Config not yet initialized!
            ============================================================

            In order to initialize your config, call 

            >>> fmdt.init(d6_dir="/your/draco6/dir", d12_dir="your/draco12/dir",
                         win_dir="/your/window/dir")

            You only ever have to do this ONE time! 
            
            For example, for a unix user ejovo we could have:

            >>> fmdt.init("/home/ejovo/Videos/Watec6mm", "/home/ejovo/Vidoes/Watec12mm",
                          "/home/ejovo/Videos/Window")

            output:

                Saved config to /home/ejovo/.local/share/fmdt_python/config.json


        """)


def init(d6_dir: str, d12_dir: str, win_dir: str):
    con = Config(d6_dir, d12_dir, win_dir)
    con.save()

def is_video(filename: str) -> bool:
    return filename[-4:] == ".avi" or filename[-4:] == ".mp4"

def is_draco6(filename: str) -> bool:
    return is_video(filename) and "Draconids-6mm" in filename

def is_draco12(filename: str) -> bool:
    return is_video(filename) and "Draconids-12mm" in filename

def is_window(filename: str) -> bool:
    return is_video(filename) and "window" in filename

def listdir() -> list[str]:
    return os.listdir(dirs.user_data_dir)

def dir() -> str:
    return dirs.user_data_dir

def setdir_draco6(d6_dir: str):
    con = load_config()
    con.d6 = d6_dir
    con.save()

def setdir_draco12(d12_dir: str):
    con = load_config()
    con.d12 = d12_dir
    con.save()

def setdir_window(win_dir: str):
    con = load_config()
    con.win = win_dir
    con.save()

def listdir_draco6() -> list[str]:
    """Return a list of videos that are found in the Watec6mm directory"""
    con = load_config()

    d6 = [v for v in os.listdir(con.d6) if is_draco6(v)]
    return d6

def listdir_draco12() -> list[str]:
    """Return a list of videos that are found in the Watec12mm directory"""
    con = load_config()

    d12 = [v for v in os.listdir(con.d12) if is_draco12(v)]
    return d12

def listdir_window() -> list[str]:
    """Return a list of videos that are in the Window directory and start with "window" """
    con = load_config()

    win = [v for v in os.listdir(con.win) if is_window(v)]
    return win 

def listdir_all() -> list[str]:

    d6 = listdir_draco6()
    d12 = listdir_draco12()
    win = listdir_window()

    all = []

    all.extend(d6)
    all.extend(d12)
    all.extend(win)

    return all

def load_gt6() -> fmdt.truth.GroundTruth:

    con = load_config()
    db = "gt6.csv"

    fmdt.download.download_draco6_csv(db)
    return fmdt.GroundTruth(con.gt6_csv(), con.d6)

def load_gt12() -> fmdt.truth.GroundTruth:

    con = load_config()
    db = "gt12.csv"

    fmdt.download.download_draco12_csv(db)
    return fmdt.GroundTruth(con.gt12_csv(), con.d12)


# if check_for_config_file():
    # con = load_config()
    
def draco6_dir():
    con = load_config()
    return con.d6

def draco12_dir():
    con = load_config()
    return con.d12

def window_dir():
    con = load_config()
    return con.win

def cache_dir():
    return dirs.user_cache_dir

def count_files_in_dir(dir: str) -> int:
    total = 0
    for _, _, files in os.walk(dir):
        total += len(files)

    return total

def size_dir(start_path = '.'):
    """Return the size of the contents of a directory in terms of bytes"""
    total_size = 0
    for dirpath, _, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size

def size_cache():
    return size_dir(cache_dir())

def bytes_format(x: int) -> str:
    """print 1024 as 1KB"""

    if abs(x) < 1024:
        return str(x) + "B"
    
    if abs(x) < 1024 * 1024:
        return str(x // 1024) + "KB"

    if abs(x) < 1024 * 1024 * 1024:
        return str(x // (1024 * 1024)) + "MB"
    
    if abs(x) < 1024 * 1024 * 1024 * 1024:
        return str(x // (1024 * 1024 * 1024)) + "GB"
    
    if abs(x) < 1024 * 1024 * 1024 * 1024 * 1024:
        return str(x // (1024 * 1024 * 1024 * 1024)) + "TB"

def clear_cache() -> int:
    """Return the total number of files and folders removed"""

    cd = cache_dir()
    paths = os.listdir(cd)
    size_cache_init = size_cache()

    files_removed = 0
    top_level_dir_removed = 0

    for p in paths:

        full_path = cd + "/" + p

        print(f"Treating file {p}...")

        # shutil.rmtree(full_path)

        if os.path.isfile(full_path):
            os.remove(full_path)
            print("file")
            files_removed += 1
            continue
    
        if os.path.isdir(full_path):
            files_removed += count_files_in_dir(p)
            print("dir")
            top_level_dir_removed += 1
            shutil.rmtree(full_path)
            continue

    print(f"{cache_dir()} cleared: {files_removed} total files and {top_level_dir_removed} top-level directories removed from cache ({bytes_format(size_cache_init - size_cache())} cleared)")


_KB = 1024
_MB = 1024 * _KB
_GB = 1024 * _MB

_CACHE_LIMIT_BYTES = 100 * _MB


def init_cache() -> None:
    if not os.path.exists(cache_dir()):
        os.mkdir(cache_dir())

    if size_cache() > _CACHE_LIMIT_BYTES:
        print(f"Cache too large! ({bytes_format(size_cache)})")
        clear_cache()

def listdir_cache() -> list[str]:
    return os.listdir(cache_dir())

def cache_info():
    print(f"Cache: {cache_dir()} has {bytes_format(size_cache())}")

def main():

    # config = Config("/home/ejovo/Videos/Watec6mm", "/home/ejovo/Videos/Watec12mm", "/home/ejovo/Videos/Window")
    # config.save()

    con = Config.load()
    print(con)

    print(listdir_draco6())
    print(listdir_draco12())
    print(listdir_window())

    print(listdir())

    gt6 = load_gt6()
    gt12 = load_gt12()

    print(gt6)
    print(gt12)

if __name__ == "__main__":
    main()


# Use different parameters of fmdt to try and find the human detected meteor 
# of draco_0
import fmdt.truth
import fmdt
import numpy as np

vid = "Draco_0.avi"

# human = fmdt.truth.HumanDetection(vid, 45, 59, 240, 226, 125, 264)

# for x in np.linspace(25, 240):
#     # print("Running fmdt.detec")
#     res = fmdt.detect(vid, light_max=255, light_min=int(x), trk_out_path="draco_tmp.txt", log=True)
#     if fmdt.truth.is_meteor_detected(human, res.tracking_list):
#         print("Human meteor detected iwth light_min:", x)

for x in [226, 231, 235]:
    # print("Running fmdt.detec")
    res = fmdt.detect(vid, light_max=255, light_min=int(x), trk_out_path="draco_tmp.txt", log=True)

    print(res.tracking_list)

vid = "draco1.avi"

for lm in np.linspace(50, 220, 10):
    res = fmdt.detect(vid, light_max=225, light_min=int(lm), trk_out_path="draco_tmp.txt", log=True)
    print(res.tracking_list)

    # if fmdt.truth.is_meteor_detected(human, res.tracking_list):
        # print("Human meteor detected iwth light_min:", x)



import fmdt
import fmdt.utils
import numpy as np
import os
import sys
import fmdt.truth

# dir = "/run/media/ejovo/Seagate Portable Drive/Meteors/Watec12mm/Meteor/"

# name = "Draconids-12mm2.00.09-1450-202900.avi"
# vid = "Draconids-6mm1.00-2750-163200.avi"; human = fmdt.truth.HumanDetection(vid, 45, 59, 240, 226, 125, 264)
vid = "draco1.avi"; human = fmdt.truth.HumanDetection(vid, 45, 59, 240, 226, 125, 264)
# vid = "Draconids-6mm1.05-0750-164200.avi"; human = fmdt.truth.HumanDetection(vid, 16, 26, 12, 339, 4, 360)
# vid = "Draconids-6mm1.14-1400-170300.avi"; human = fmdt.truth.HumanDetection(vid, 11, 24, 11, 233, 14, 273)
# vid = "Draconids-6mm1.20-2350-171600.avi" ; human = fmdt.truth.HumanDetection(vid, 22, 38, 232, 368, 237, 396)

# vid = dir + name

if vid is None and len(sys.argv) < 2:
    assert False, "No video selected. Pass the path of a video as an argument or change the variable `vid` in this script"

human.test_detection_vary_light(vid, offset=25, light_min_start=150, light_min_end=235, k_trials=15, log=True)



"""Convert the contents of `human_detectsions.csv` and `test_vid.csv` and "meteors.txt" into a single data base file `videos.db`"""

import sqlite3 
import pandas as pd
from fmdt.args import *
import fmdt.truth

f = open("videos.db", "w")
f.close()

con = sqlite3.connect("videos.db")

cur = con.cursor()

cur.execute("CREATE TABLE video(id, name, type)")

# Let's load in our csv
df = pd.read_csv("test_vid.csv")
print(df.head(10))

for i in range(len(df)):
    r = df.iloc[i]

    id = r["id"]
    name = r["name"]
    type = r["type"]


    # print(id)
    # print(name)
    # print(type)
    cur.execute(f"""
    INSERT INTO video VALUES 
        ({id}, '{name}', '{type}')
    """)

#================ GroundTruths ===================
# fmdt.truth.GroundTruth("")
# fmdt.truth.init_ground_truth()
csv_file = "human_detections.csv"

df_gt = pd.read_csv(csv_file)

# Pandas DataFrame Ground Truths
cur.execute("CREATE TABLE human_detection(id, video_name, start_x, start_y, start_frame, end_x, end_y, end_frame)")
for i in range(len(df_gt)):
    r = df_gt.iloc[i]

    vid_name = r["video_name"]
    start_x = r["start_x"]
    start_y = r["start_y"]
    start_frame = r["start_frame"]
    end_x = r["end_x"]
    end_y = r["end_y"]
    end_frame = r["end_frame"]

    cur.execute(f"""
    INSERT INTO human_detection VALUES 
        ({i}, '{vid_name}', {start_x}, {start_y}, {start_frame}, {end_x}, {end_y}, {end_frame})
    """)

print(i)

import fmdt.db
import fmdt.truth

# ================= Ground Truths in meteors.txt ========
meteors = fmdt.truth.load_meteors_file("meteors.txt", "2022_05_31_tauh_34_meteors.mp4")

start = i + 1
for i in range(start, start + len(meteors)):
    m = meteors[i - start]

    query = f"""
    INSERT INTO human_detection VALUES 
        ({i}, '{m.video_name}', {m.start_x}, {m.start_y}, {m.start_frame}, {m.end_x}, {m.end_y}, {m.end_frame})
    """

    cur.execute(query)

con.commit()
con.close()




"""Arguments Class to store shared parameters when calling a chain of .detect(args).visu().split()"""
import fmdt.api
# import fmdt.res
import fmdt.core
import fmdt.utils
import shutil
import subprocess
import pandas as pd
from enum import Enum
import pickle
import hashlib
import copy


class Result:
    pass
    

def filter_dict(d: dict):
    """Filter out None values in a dict, returning a new dict"""
    out = {}

    for (k, v) in d.items():
        if not v is None:
            out[k] = v

    return out

def row_to_dict(row: pd.Series) -> dict:
    """Convert the non Na values of a pd.DataFrame row to a dict"""
    out = {}
    for k in row.keys():
        if not pd.isna(row[k]):
            out[k] = row[k]
    
    return out
    
class DetectArgs:

    def __init__( 
        self,
        vid_in_path: str, 
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = None,
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        trk_out_path: str | None = None,
    ):

        self.vid_in_path = vid_in_path
        self.vid_in_start = vid_in_start
        self.vid_in_stop = vid_in_stop
        self.vid_in_skip = vid_in_skip
        self.vid_in_buff = vid_in_buff
        self.vid_in_loop = vid_in_loop
        self.vid_in_threads = vid_in_threads
        self.light_min = light_min 
        self.light_max = light_max 
        self.ccl_fra_path = ccl_fra_path
        self.ccl_fra_id = ccl_fra_id
        self.mrp_s_min = mrp_s_min
        self.mrp_s_max = mrp_s_max
        self.knn_k = knn_k
        self.knn_d = knn_d
        self.knn_s = knn_s
        self.trk_ext_d = trk_ext_d
        self.trk_ext_o = trk_ext_o
        self.trk_angle = trk_angle
        self.trk_star_min = trk_star_min
        self.trk_meteor_min = trk_meteor_min
        self.trk_meteor_max = trk_meteor_max
        self.trk_ddev = trk_ddev
        self.trk_all = trk_all
        self.trk_bb_path = trk_bb_path
        self.trk_mag_path = trk_mag_path
        self.trk_out_path = trk_out_path
        self.log_path = log_path

    def to_dict(self) -> dict:
        return {
            "vid_in_path": self.vid_in_path, 
            "vid_in_start": self.vid_in_start,
            "vid_in_stop": self.vid_in_stop,
            "vid_in_skip": self.vid_in_skip,
            "vid_in_buff": self.vid_in_buff,
            "vid_in_loop": self.vid_in_loop,
            "vid_in_threads": self.vid_in_threads,
            "light_min": self.light_min,
            "light_max": self.light_max,
            "ccl_fra_path": self.ccl_fra_path,
            "ccl_fra_id": self.ccl_fra_id,
            "mrp_s_min": self.mrp_s_min,
            "mrp_s_max": self.mrp_s_max,
            "knn_k": self.knn_k,
            "knn_d": self.knn_d,
            "knn_s": self.knn_s,
            "trk_ext_d": self.trk_ext_d,
            "trk_ext_o": self.trk_ext_o,
            "trk_angle": self.trk_angle,
            "trk_star_min": self.trk_star_min,
            "trk_meteor_min": self.trk_meteor_min,
            "trk_meteor_max": self.trk_meteor_max,
            "trk_ddev": self.trk_ddev,
            "trk_all": self.trk_all,
            "trk_bb_path": self.trk_bb_path,
            "trk_mag_path": self.trk_mag_path,
            "log_path": self.log_path,
            "trk_out_path": self.trk_out_path 
        }
    
    def to_reduced_dict(self) -> dict:
        d = self.to_dict()
        out = {}
        for (k, v) in d.items():
            if not v is None:
                out[k] = v
        
        return out
    
    def argv(self) -> list[str]:
        """Return a list of arguments that will be used to execute fmdt-detect"""
        return handle_detect_args(**self.to_dict())

    def cmd(self) -> str:
        return ' '.join(self.argv())
    
    def exec(self, log: bool = False, timeout: float = None, cache: bool = False, save_df: bool = False):
        res = fmdt.api.detect(**self.to_dict(), log=log, timeout=timeout, cache=cache, save_df=save_df)
        return res
    
    def strip(self):
        """Strip Args of path variables that don't affect execution"""

        c = copy.deepcopy(self)
        c.ccl_fra_path = None
        c.log_path = None
        c.trk_bb_path = None
        c.trk_mag_path = None
        c.trk_out_path = None

        return c
        

    def to_bytes(self) -> bytes:
        """Convert to bytes, removing influence of all path variables"""
        return pickle.dumps(self.strip())
    
    def digest(self) -> str:
        return hashlib.md5(self.to_bytes()).hexdigest()

    def __hash__(self) -> int:
        # return int.from_bytes(pickle.dumps(self), "big")
        return int(self.digest, 16)
    
    def gen_unique_dir(self) -> str:
        return self.digest()[0:16] 
    
    def cache_dir(self) -> str:
        return fmdt.cache_dir() + "/" + self.gen_unique_dir()

    def cache_trk(self) -> str:
        """Generate the full path to a unique file to store the results of this detection"""
        return self.cache_dir() + "_trk.txt"

class VisuArgs:

    def __init__(
            self, 
            vid_in_path: str, 
            trk_path: str,
            trk_bb_path: str,
            vid_out_path: str,
            vid_in_start: int | None = None, 
            vid_in_stop: int | None = None, 
            vid_in_threads: int | None = None, 
            trk_id: bool = False,
            trk_nat_num: bool = False,
            trk_only_meteor: bool = False,
            gt_path: str | None = None
        ):

        self.vid_in_path = vid_in_path
        self.vid_in_start = vid_in_start
        self.vid_in_stop = vid_in_stop
        self.vid_in_threads = vid_in_threads
        self.trk_path = trk_path
        self.trk_bb_path = trk_bb_path
        self.trk_id = trk_id
        self.trk_nat_num = trk_nat_num
        self.trk_only_meteor = trk_only_meteor
        self.gt_path = gt_path
        self.vid_out_path = vid_out_path

    def to_dict(self) -> dict:
        return {

            "vid_in_path": self.vid_in_path,
            "vid_in_start": self.vid_in_start,
            "vid_in_stop": self.vid_in_stop,
            "vid_in_threads": self.vid_in_threads,
            "trk_path": self.trk_path,
            "trk_bb_path": self.trk_bb_path,
            "trk_id": self.trk_id,
            "trk_nat_num": self.trk_nat_num,
            "trk_only_meteor": self.trk_only_meteor,
            "gt_path": self.gt_path,
            "vid_out_path": self.vid_out_path
        }
    
    def to_reduced_dict(self) -> dict:
        d = self.to_dict()
        out = {}
        for (k, v) in d.items():
            if not v is None:
                out[k] = v
        
        return out
    
    def argv(self) -> list[str]:
        """Return a list of arguments that will be used to execute fmdt-detect"""
        return handle_visu_args(**self.to_dict())

    def cmd(self) -> str:
        return ' '.join(self.argv())
    
    def exec(self, log: bool = False, timeout: float = None):
        res = fmdt.api.visu(**self.to_dict())
        return res




class Args:
    """Args keeps track of a configuration of parameters for all of fmdt's executables

    Upon a call to `fmdt.detect()`, an Args object is returned that remembers
    the configuration used to execute fmdt-detect. We can subsequently use the
    Args object as an interface to directly calling fmdt-visu, without having to respecify some of the more specific arguments.

    Consider the difference between the two calls:
    ```
    fmdt.detect(vid_in_path = "vid.mp4")
    fmdt.visu(vid_in_path = "vid.mp4", vid_out_path = "vid_visu.mp4")
    ```

    and
    ```
    fmdt.detect(vid_in_path = "vid.mp4").visu()
    ```  
    where storing the configuration for `fmdt-detect` in an Args object allows us to fill
    in the parameters needed by a call to `fmdt-visu`.

    This class is therefore a collection of dictionaries of parameters and interfaces to
    the fmdt.api functions.

    ==================

    Args has the fields

    detect_args: DetectArgs
    visu_args: VisuArgs
    log: bool
    timeout: float
    """

    def __init__(
        self,
        detect_args: DetectArgs,
        visu_args: VisuArgs,
        log: bool = False,
        timeout: float | None = None
    ):
        self.detect_args = detect_args
        self.visu_args = visu_args
        self.log = log
        self.timeout = timeout

    @staticmethod
    def new(
        #=================== DetectArgs =============================
        vid_in_path: str = "default.mp4", 
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = None,
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        trk_out_path: str | None = None,
        #=================== VisuArgs =================================
        trk_id: bool | None = None,
        trk_nat_num: bool| None = None,
        trk_only_meteor: bool | None = None,
        gt_path: str | None = None,
        vid_out_path: str | None = None,
        #================== PythonArgs================================
        log: bool | None = False, 
        timeout: float | None = None,
    ):

        detect_args = DetectArgs(vid_in_path=vid_in_path,
                              vid_in_start=vid_in_start,
                              vid_in_stop=vid_in_stop,
                              vid_in_skip=vid_in_skip,
                              vid_in_buff=vid_in_buff,
                              vid_in_loop=vid_in_loop,
                              vid_in_threads=vid_in_threads,
                              light_min =light_min ,
                              light_max =light_max ,
                              ccl_fra_path=ccl_fra_path,
                              ccl_fra_id=ccl_fra_id,
                              mrp_s_min=mrp_s_min,
                              mrp_s_max=mrp_s_max,
                              knn_k=knn_k,
                              knn_d=knn_d,
                              knn_s=knn_s,
                              trk_ext_d=trk_ext_d,
                              trk_ext_o=trk_ext_o,
                              trk_angle=trk_angle,
                              trk_star_min=trk_star_min,
                              trk_meteor_min=trk_meteor_min,
                              trk_meteor_max=trk_meteor_max,
                              trk_ddev=trk_ddev,
                              trk_all=trk_all,
                              trk_bb_path=trk_bb_path,
                              trk_mag_path=trk_mag_path,
                              trk_out_path=trk_out_path,
                              log_path=log_path)
        

        if vid_out_path is None:
            name, ext = fmdt.utils.decompose_video_filename(vid_in_path)
            vid_out_path = f"{name}_visu.{ext}"

        visu_args = VisuArgs(vid_in_path=vid_in_path,
                            vid_in_start=vid_in_start,
                            vid_in_stop=vid_in_stop,
                            vid_in_threads=vid_in_threads,
                            trk_path=trk_out_path,
                            trk_bb_path=trk_bb_path,
                            trk_id=trk_id,
                            trk_nat_num=trk_nat_num,
                            trk_only_meteor=trk_only_meteor,
                            gt_path=gt_path,
                            vid_out_path=vid_out_path)
        
        return Args(detect_args, visu_args, log, timeout) 

    def __str__(self) -> str:

        a0 = "<fmdt.args.Args object>"
        a1 = "\n===================="
        a2 = "\nDetect parameters: \n"
        a = str(self.detect_args.to_reduced_dict())

        return a0 + a1 + a2 + a


    # We should change this to print out only the arguments that are not none
    # def __str__(self) -> str:
    #     if not self.detect_args:
    #         f"detect_args:"

    # def __repr__(self) -> str:

    #     if not self.detect_args is None:
    #         d = filter_dict(self.detect_args)

    #     # s = ""
    #     # for (k, v) in self.detect_args.items():
    #         # if not v is None:
    #             # s += f'"{k}": {v}, ' 
        
    #     # return "{" + s[:-2] + "}"
    #     return d.__repr__()
    
    def detect(self, cache: bool = False, save_df: bool = False):
        """OOP Interface to calling fmdt.api.detect()
        
        Parameters
        ----------
        cache (bool): When true, store the output of fmdt-detect in a unique 
            file corresponding to the set of parameters 
        """

        # Make sure the detecting arguments are not none
        if self.detect_args is None:
            self.detect_args = DetectArgs(**default_detect_args())
        
        return self.detect_args.exec(self.log, self.timeout, cache, save_df)
    
    def does_detect_fail(self, log=False) -> bool:
        """Returns true if the stderr pipe of a call to fmdt-detect is not empty"""

        if self.detect_args["vid_in_path"] is None:
            return True

        detect_cmd = detect_args_to_cmd(self.detect_args)
        if (log):
            print(detect_cmd)
        res = subprocess.run(detect_cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        # print(res.stderr.decode())

        if res.returncode != 0:
            return True
        
        return False
    
    def detect_cmd(self) -> str:
        return handle_detect_args(**self.detect_args)

    def visu(self, **kwargs):
        """OOP Interface to calling fmdt.api.visu()"""

        # Do we have visu arguments?
        assert not self.visu_args is None, "No visu args for this object"

        self.visu_args.exec() 

        return self
    
    def vid(self) -> str | None:
        """Return the name of the video, if it exists"""
        if not self.detect_args.vid_in_path is None:
            return self.detect_args.vid_in_path
        elif not self.visu_args.vid_in_path is None:
            return self.visu_args.vid_in_path
        else:
            return None
        
    def tracks(self) -> str | None:
        """Return the name of the tracks file, if it exists"""
        if not self.detect_args.trk_out_path is None:
            return self.detect_args.trk_out_path
        elif not self.visu_args.trk_path is None:
            return self.visu_args.trk_path
        else:
            return None

    def bbs(self) -> str | None:
        """Return the name of the bounding boxes (BBs) file, if it exists"""
        if not self.detect_args.trk_bb_path is None:
            return self.detect_args.trk_bb_path
        elif not self.visu_args.trk_bb_path is None:
            return self.visu_args.trk_bb_path
        else:
            return None

    def split(self, nframes_buffer=3, overwrite=False, exact_split=False, log=False):
        # Check to see if we have the tracks saved.
        fmdt.core.split_video_at_meteors(video_filename=self.vid(), 
                                         detect_tracks_in=self.tracks(),
                                         nframes_before=nframes_buffer,
                                         nframes_after=nframes_buffer,
                                         overwrite=overwrite,
                                         exact_split=exact_split,
                                         log=log)

    # Take the detect argument dictionary and write out a comma separated value string
    def detect_csv_header(self) -> str:
        """Return the header line used in a csv file storing multiple argument configurations"""
        header = ""
        for k in self.detect_args.keys():
            if k == "log":
                continue
            header = header + f"{k},"

        return header[:-1] + "\n"

    def detect_to_csv_row(self) -> str:
        """Return the csv representation of a detect_args dict"""
        csv = ""
        for v in self.detect_args.values():
            if v is None:
                csv = csv + ","
            else:
                csv = csv + f"{str(v)},"
        
        # Drop the last comma
        return csv[:-1] + "\n"
    
    def get_tracking_list(self) -> list[fmdt.core.TrackedObject]:
        """Retreive the list of TrackedObject that is stored in the trk_out_path file"""
        assert not self.detect_args.trk_out_path is None, "Out track file not stored"

        return fmdt.core.extract_all_information(self.detect_args.trk_out_path)
    
    def command(self) -> str:
        """Return the command used to execute fmdt-detect with this configuration"""
        return ' '.join(handle_detect_args(**self.detect_args))
    
    def digest(self) -> str:
        return hashlib.md5(pickle.dumps(self)).hexdigest()

    def __hash__(self) -> int:
        # return int.from_bytes(pickle.dumps(self), "big")
        return int(hashlib.md5(pickle.dumps(self)).hexdigest(), 16)
    
    def gen_unique_trk(self) -> str:
        """Generate a unique trk file corresponding to this set of parameters"""

        h = self.digest()
        return "trk_" + h[0:16] + ".txt" 

    
    # Write the dictionary of fmdt-detect arguments to a csv file
    # def detect_to_csv(self, csv_filename) -> None:

class FMDTResult:

    def __init__(self, arg_err: list[float] = None, ecarts_type: list[float] = None, nrois: list[int] = None, args: Args = None):
        self.arg_err = arg_err
        self.ecarts_type = ecarts_type
        self.nrois = nrois
        self.args = args

# Need some functions to retrieve the results from a log thing






def list_args_to_csv(lst: list[Args], csv_file: str):

    with open(csv_file, "w") as f:

        f.write(lst[0].detect_csv_header())
        for a in lst:
            f.write(a.detect_to_csv_row())

def csv_to_list_args(csv_file: str) -> list[Args]:

    df = pd.read_csv(csv_file)
    n_rows = len(df)
    
    return [Args.from_row(df.loc[i]) for i in range(n_rows)]

def video_input(filename: str) -> Args:
    a = Args()
    a.detect_args = default_detect_args()
    a.detect_args.vid_in_path = filename
    return a


# Create an Args object from the fmdt-detect parameters
def detect_args(
        vid_in_path: str = "default.mp4", 
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = None,
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        trk_out_path: str | None = None,
        log: bool = False,
        timeout: float = None
    ) -> Args:
    """Convert the parameters used in fmdt.detect into an Args object"""

    assert not vid_in_path is None, "vid_in_path cannot be None"

    d_args = DetectArgs(vid_in_path, vid_in_start, vid_in_stop, vid_in_skip, vid_in_buff,
                        vid_in_loop, vid_in_threads, light_min, light_max, ccl_fra_path,
                        ccl_fra_id, mrp_s_min, mrp_s_max, knn_k, knn_d, knn_s, trk_ext_d,
                        trk_ext_o, trk_angle, trk_star_min, trk_meteor_min, trk_meteor_max,
                        trk_ddev, trk_all, trk_bb_path, trk_mag_path, log_path, trk_out_path)
    
    name, ext = fmdt.utils.decompose_video_filename(vid_in_path)
    visu_name = f"{name}_visu.{ext}"

    v_args = VisuArgs(vid_in_path=vid_in_path,
                      vid_in_start=vid_in_start,
                      vid_in_stop=vid_in_stop,
                      vid_in_threads=vid_in_threads,
                      trk_path=trk_out_path,
                      trk_bb_path=trk_bb_path,
                      vid_out_path=visu_name)
    
    return Args(d_args, v_args, log, timeout)

def visu_args(
        vid_in_path: str = None,
        vid_in_start: int = None,
        vid_in_stop: int = None,
        vid_in_threads: int = None,
        trk_path: str = None,
        trk_bb_path: str = None,
        trk_id: bool = None,
        trk_nat_num: bool = None,
        trk_only_meteor: bool = None,
        gt_path: str = None,
        vid_out_path: str = None
    ) -> VisuArgs:

    # wtf man
    return VisuArgs( **{
        "vid_in_path": vid_in_path,
        "vid_in_start": vid_in_start,
        "vid_in_stop": vid_in_stop,
        "vid_in_threads": vid_in_threads,
        "trk_path": trk_path,
        "trk_bb_path": trk_bb_path,
        "trk_id": trk_id,
        "trk_nat_num": trk_nat_num,
        "trk_only_meteor": trk_only_meteor,
        "gt_path": gt_path,
        "vid_out_path": vid_out_path
    }
    )

def default_detect_args() -> dict:       
    # Hi 
    default_detect = {
        "vid_in_path": None, 
        "vid_in_start": None,
        "vid_in_stop": None,
        "vid_in_skip": None,
        "vid_in_buff": None,
        "vid_in_loop": None,
        "vid_in_threads": None,
        "light_min": None,
        "light_max": None,
        "ccl_fra_path": None,
        "ccl_fra_id": None,
        "mrp_s_min": None,
        "mrp_s_max": None,
        "knn_k": None,
        "knn_d": None,
        "knn_s": None,
        "trk_ext_d": None,
        "trk_ext_o": None,
        "trk_angle": None,
        "trk_star_min": None,
        "trk_meteor_min": None,
        "trk_meteor_max": None,
        "trk_ddev": None,
        "trk_all": None,
        "trk_bb_path": None,
        "trk_mag_path": None,
        "log_path": None,
        "trk_out_path": None
    }

    return default_detect

def handle_detect_args(
        vid_in_path: str, 
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = None,
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        **args
    ) -> list[str]:
    """Convert the arguments needed for fmdt-detect into a fmdt-detect command-line call

        
    """

    fmdt_detect_exe = shutil.which("fmdt-detect")
    fmdt_detect_found = not fmdt_detect_exe is None
    assert fmdt_detect_found, "fmdt-detect executable not found"
    args = [fmdt_detect_exe, "--vid-in-path", vid_in_path]

    def arg_str(arg: str | None, flag: str) -> None:
        """Modify args in place if arg is not None"""
        if not arg is None:
            args.extend(["--" + flag, str(arg)])

    def arg_bool(arg: bool | None, flag: str) -> None:
        """Modify args in place if arg is True"""
        if arg:
            args.append("--" + flag)

    # ====== Arguments of the form --flag <arg_value>
    arg_str(vid_in_skip, "vid-in-skip")
    arg_str(vid_in_threads, "vid-in-threads")
    arg_str(light_min, "ccl-hyst-lo")
    arg_str(light_max, "ccl-hyst-hi")
    arg_str(ccl_fra_path, "ccl_fra_path")
    arg_str(mrp_s_min, "mrp-s-min")
    arg_str(mrp_s_max, "mrp-s-max")
    arg_str(knn_k, "knn-k")
    arg_str(knn_d, "knn-d")
    arg_str(knn_s, "knn-s")
    arg_str(trk_ext_d, "trk-ext-d")
    arg_str(trk_ext_o, "trk-ext-o")
    arg_str(trk_star_min, "trk-star-min")
    arg_str(trk_meteor_min, "trk-meteor-min")
    arg_str(trk_meteor_max, "trk-meteor-max")
    arg_str(trk_ddev, "trk-ddev")
    arg_str(trk_mag_path, "trk-mag-path")
    arg_str(log_path, "log-path")
    arg_str(trk_angle, "trk-angle")
    arg_str(trk_star_min, "trk-star-min")
    arg_str(vid_in_start, "vid-in-start")
    arg_str(vid_in_stop, "vid-in-stop")
    arg_str(trk_bb_path, "trk-bb-path")
    arg_str(vid_in_loop, "vid-in-loop")

    # ======== Arguments of the form --toggle
    arg_bool(vid_in_buff, "vid_in_buff")
    arg_bool(ccl_fra_id, "ccl-fra-id")
    arg_bool(trk_all, "trk-all")

    return args

def handle_visu_args(
        vid_in_path: str = None,
        vid_in_start: int = None,
        vid_in_stop: int = None,
        vid_in_threads: int = None,
        trk_path: str = None,
        trk_bb_path: str = None,
        trk_id: bool = None,
        trk_nat_num: bool = None,
        trk_only_meteor: bool = None,
        gt_path: str = None,
        vid_out_path: str = None
    ) -> list[str]:


    fmdt_visu_exe = shutil.which("fmdt-visu")
    fmdt_visu_found = not fmdt_visu_exe is None
    assert fmdt_visu_found, "fmdt-visu executable not found"


    assert not vid_in_path is None, "No input video specified"
    assert not trk_path is None, "No input track path"
    assert not trk_bb_path is None, "No input bounding boxes file"

    args = [fmdt_visu_exe, "--vid-in-path", vid_in_path, "--trk-bb-path", trk_bb_path, "--trk-path", trk_path] 

    if not vid_out_path is None:
        args.extend(["--vid-out-path", vid_out_path])
    else:
        name, ext = fmdt.utils.decompose_video_filename(vid_in_path)
        new_name = f"{name}_visu.{ext}"
        args.extend(["--vid-out-path", new_name])

    # helper closure to clean up repetitive code
    def add_arg(arg, flag):
        if not arg is None:
            args.extend([flag, str(args)])

    add_arg(vid_in_start, "--vid-in-start")
    add_arg(vid_in_stop, "--vid-in-stop")
    add_arg(vid_in_threads, "--vid-in-threads")
    add_arg(gt_path, "--gt-path")

    if trk_id:
        args.append("--trk-id")

    if trk_nat_num:
        args.append("--trk-nat-num")

    if trk_only_meteor:
        args.append("--trk-only-meteor")

    return args



def detect_args_to_cmd(args: dict) -> list[str]:
    return handle_detect_args(**args)            

    
def main() -> None:
    a = Args()
    # a.detect_args = default_detect_args()
    a.detect_args = detect_args(vid_in_path="demo.mp4")
    print(a.detect_csv_header())
    print(a.detect_to_csv_row())


"""Load in a data base stored in DATA_BASE_FILE


"""
import fmdt
import fmdt.truth
import fmdt.args
# import pandas as pd
from os import listdir


DATA_BASE_FILE = "human_detections.csv"
# DATA_BASE_FILE = "human_detection.csv"

humans = fmdt.HumanDetection.init_ground_truth(DATA_BASE_FILE)

# file = "Draconids-6mm1.14-1400-170300.avi"

# meteors_in_file = [m for m in humans if m.video_name == file]

# print(humans)

# for h in humans:
#     print(h)

# print(meteors_in_file)
# for m in meteors_in_file:
#     print(m)

# Now specify a directory that we want to test against our database
watec6  = "/home/ejovo/Videos/Watec6mm"
watec12 = "/home/ejovo/Videos/Watec12mm"
# dir = watec12
dir = watec6

is_video = lambda x : x[-3:] == "avi" or x[-3:] == "mp4"
entries = listdir(dir)
entries = [v for v in entries if is_video(v)]

def get_meteors(vid: str) -> bool:
    return [m for m in humans if m.video_name == vid]

def has_meteors(vid: str) -> bool:
    return len(get_meteors(vid)) > 0

ground_truths = [get_meteors(video) for video in entries if has_meteors(video)]

print(f"Detected {len(ground_truths)} videos with ground truth meteors")

successful_args = []

MAX_ITER = 100 
i = 0

for truth_list in ground_truths: 
    # most times theres only one element in this list

    if i >= MAX_ITER:
        break
    
    i += 1
    args = []

    for truth in truth_list:
        full_vid_name = dir + "/" + truth.video_name
        print(f"Processing video: {full_vid_name}")

        args.append(truth.test_detection_vary_light(full_vid_name,
                                                    k_trials = 10))
    
    successful_args.append(args)

print(successful_args)

flattened_args = []

for video in successful_args:
    for meteor in video:
        for args in meteor:
            # print(f"args: {args} detected meteor!")
            flattened_args.append(args)
        
for a in flattened_args:
    print(a)

fmdt.args.list_args_to_csv(flattened_args, "database_6mm.csv")
# fmdt.args.list_args_to_csv(flattened_args, "database_12mm.csv")


"""Load up a configuration of fmdt-detect commands to run, and expect to find a specific meteor"""

import fmdt.truth
import fmdt

import random

DATA_BASE_FILE = "human_detections.csv"
# DATA_BASE_FILE = "human_detection.csv"

humans = fmdt.HumanDetection.init_ground_truth(DATA_BASE_FILE)

# watec6 = None
# watec12 = None
watec6  = "/home/ejovo/Videos/Watec6mm"
watec12 = "/home/ejovo/Videos/Watec12mm"

# I actually think that I might want to 

# Need to load in the configs 
# Load them in as Args
gt12 = fmdt.init_ground_truth(vid_db_dir=watec12)
gt12.meteors = [m for m in gt12.meteors if m.is_draco_12()]


gt6 = fmdt.init_ground_truth(vid_db_dir=watec6)
gt6.meteors = [m for m in gt6.meteors if m.is_draco_6()]

print(gt6)
print(gt12)

print(gt6.meteors)

print(gt6.vids())


d_args = {
    "light_min": 221,
    "light_max": 246,
    "trk_bb_path": "bb.txt",
    "trk_out_path": "trk.txt",
    "timeout": 1
}

a = fmdt.Args(detect_args=d_args)

# Apply a single command across the entire database
# detections = gt6.try_command(a)
# print(detections)


# EXPERIMENTAL RESULTS
detections = [False, False, True, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, True, False, False, False]


print(f"Sum detections: {sum(detections)}")
print(gt6)

def roll_parameters():
    light_min = random.randint(0, 254)
    diff = random.randint(1, 255 - light_min)

    d_args["light_min"] = light_min
    d_args["light_max"] = light_min + diff

    det = gt6.try_command(fmdt.Args(detect_args=d_args))
    print(f"[{light_min}, {light_min + diff}] detected {sum(det)} gts")

# roll_parameters()


N_TRIALS = 05 
# N_TRIALS = 150  

# I want a list of light min and light max pairs
lm_pairs = []
gts_det  = []
lst_detections = [[]]

MAX_LIGHT_DIFF = 40
LIGHT_MIN_MIN = 150
LIGHT_MIN_MAX = 254

# ============================ ~ 20 minute run ========================
for k in range(N_TRIALS):
    light_min = random.randint(LIGHT_MIN_MIN, LIGHT_MIN_MAX)
    diff = random.randint(1, min(255 - light_min, MAX_LIGHT_DIFF))

    d_args["light_min"] = light_min
    d_args["light_max"] = light_min + diff

    # det = gt6.try_command(fmdt.Args(detect_args=d_args))
    det = gt12.try_command(fmdt.Args(detect_args=d_args))
    print(f"[{light_min}, {light_min + diff}] detected {sum(det)} gts")

    lm_pairs.append((light_min, light_min + diff))
    gts_det.append(sum(det))
    lst_detections.append(det)

# print(lm_pairs)
# print(gts_det)
# ======================================================================

# #! =============== Experimental Results ===========================
# lm_pairs = [(253, 254), (164, 176), (215, 237), (200, 214), (218, 230), (250, 253), (186, 188), (250, 255), (199, 238), (151, 190), (220, 230), (202, 214), (191, 211), (205, 227), (224, 252), (250, 253), (227, 237), (196, 235), (176, 189), (194, 219)]
# gts_det = [8, 0, 2, 0, 3, 6, 2, 6, 2, 0, 3, 1, 1, 0, 3, 6, 3, 1, 0, 0]

# # Let's try and plot this as a 3d surface?

# import matplotlib.pyplot as plt
# import matplotlib.patches as patches
# import numpy as np

lmin_lst = [l[0] for l in lm_pairs]
lmax_lst = [l[1] for l in lm_pairs]

print(lmin_lst)
print(lmax_lst)
print(gts_det)





"""Class to handle the ground truth of a video

HumanDetections are loaded in with a video name that is completely
independent of the path. Users can set the fmdt variables WATEC6_DIR
and WATEC12_DIR to execute database tests
"""

import pandas as pd
import math
import fmdt.core
import fmdt.utils
import fmdt.args
from fmdt.core import TrackedObject
import numpy as np
import os
from termcolor import colored
from deprecated import deprecated

WATEC6_DIR:  str = "./"
WATEC12_DIR: str = "./"

class HumanDetection:

    GROUND_TRUTH = None

    def __init__(
            self,
            video_name: str,
            start_frame: int,
            end_frame: int,
            start_x: float,
            start_y: float,
            end_x: float,
            end_y: float
        ):

        self.video_name = video_name
        self.start_frame = start_frame
        self.end_frame = end_frame
        self.start_x = start_x
        self.start_y = start_y
        self.end_x = end_x
        self.end_y = end_y

    def __str__(self) -> str:
        return (
            f"""({self.video_name}, f0: {self.start_frame}, fT: {self.end_frame}, pos0: ({self.start_x}, {self.start_y}), posT: ({self.end_x}, {self.end_y}))"""
        )

    def __lt__(self, other):
        return self.video_name < other.video_name
    
    def is_draco_6(self) -> bool:
        """Return true if part of the Draconids6mm series"""
        # vid_name = self.video_name.split("/")[-1]
        v, _ = fmdt.utils.decompose_video_filename(self.video_name)
        return "Draconids-6mm" in v 
    
    def is_draco_12(self) -> bool:
        """Return true if part of the Draconids6mm series"""
        # vid_name = self.video_name.split("/")[-1]
        v, _ = fmdt.utils.decompose_video_filename(self.video_name)
        return "Draconids-12mm" in v 
    
    def delta_x(self) -> float:
        """Displacement in the x_axis direction"""
        return self.end_x - self.start_x
    
    def delta_y(self) -> float:
        """Displacement in the y direction"""
        return self.end_y - self.start_y
    
    def displacement(self) -> tuple[float, float]:
        return (self.delta_x(), self.delta_y())
    
    def lifetime(self) -> tuple[int, int]:
        return (self.start_frame, self.end_frame)
    
    def nframes_alive(self) -> int:
        return self.end_frame - self.start_frame

    def slope(self) -> float:
        if self.delta_x() == 0:
            return float("inf")
        return self.delta_y() / self.delta_x()
    
    def direction(self) -> float:
        """Return the angle of the displacement vector of this meteor. Units are radians"""
        return math.atan2(self.delta_y(), self.delta_x())
    
    def dx(self) -> float:
        return self.delta_x() / self.nframes_alive()
    
    def interpolate_pos(self, frame_n: int) -> tuple[float, float]:
        f_prime = frame_n - self.start_frame
        xi = self.start_x + (self.dx() * f_prime)
        yi = self.start_y + (self.dx() * self.slope()) * f_prime
        return (xi, yi)

    def is_detected_in_list(self, tracking_list: list[TrackedObject]) -> bool:

        # keep only meteors
        # detected_m = [m for m in tracking_list if m.is_meteor()]

        for tracked in tracking_list: 
            if are_objects_the_same(self, tracked):
                return True

        return False
    
    def test_detection_vary_light(
            self,
            vid: str,
            offset: int = 25,
            light_min_start: int = 150,
            light_min_end: int = 230,
            k_trials: int = 10,
            log: bool = False
        ):

        return vary_light_intervals(
            vid=vid,
            truth=self,
            offset=offset,
            light_min_start=light_min_start,
            light_min_end=light_min_end,
            k_trials=k_trials,
            log=log
        )
    

    def detect(self, args: fmdt.args.Args) -> bool:
        """Paired up with an Args object, frun with fmdt-detect and 
        see if this Truth is registered"""
        res = args.detect()
        return is_meteor_detected(self, res.tracking_list)

    def to_alsoc_format(self) -> str:
        """Convert a HumanDetection into a string specified by the format:

            "{otype} {fbeg} {xbeg} {ybeg} {fend} {xend} {yend}\n"

        """
        return f"meteor {self.start_frame} {self.start_x} {self.start_y} {self.end_frame} {self.end_x} {self.end_y}\n"

    @staticmethod
    def from_pd_row(row: pd.Series, dir_prepend = ""):
        # if dir_prepend[-1] != "/":
            # dir_prepend = dir_prepend + "/"
        return HumanDetection(dir_prepend + row["video_name"],
                            int(row["start_frame"]),
                            int(row["end_frame"]),
                            float(row["start_x"]),
                            float(row["start_y"]),
                            float(row["end_x"]),
                            float(row["end_y"]))

    @staticmethod
    def from_alsoc_format(video_name, alsoc: str):

        lst = alsoc.strip().split()

        return HumanDetection(video_name, start_frame=int(lst[1]), start_x=float(lst[2]), start_y=float(lst[3]),
                              end_frame=int(lst[4]), end_x=float(lst[5]), end_y = float(lst[6]))



    @deprecated(version="0.25.0", reason="We should load in our GroundTruth's with the fmdt.load_gt6() and fmdt.load_gt12() functions")
    @staticmethod
    def init_ground_truth(database_filename: str, video_db_dir: str = "./"):
        """"""
        HumanDetection.GROUND_TRUTH = read_human_detection_csv(database_filename, video_db_dir) 
        return read_human_detection_csv(database_filename, video_db_dir)
    
def load_meteors_file(filename: str, video_name: str) -> list[HumanDetection]:
    """Load in a meteors file (ex: https://github.com/alsoc/fmdt/blob/develop/validation/2022_05_31_tauh_34_meteors.txt)
    as a list of `HumanDetection`
    
    Parameters
    ----------

    filename (str): path to the containing the ground truths in alsoc format
    video_name (str): name of the video that these ground truths are associated with
    """

    with open(filename) as file:
        lines = file.readlines()

        out = [HumanDetection.from_alsoc_format(video_name=video_name, alsoc=l) for l in lines]

    return out

def save_meteors_file(filename: str, meteors: list[HumanDetection]):

    with open(filename, "w") as file:

        for m in meteors:
            file.write(m.to_alsoc_format())


# Verite de terrains for a bunch of individually tracked meteors
class GroundTruth:

    def __init__(self, csv: str, vid_dir: str = './'):
        self.vid_dir = vid_dir

        self.meteors = read_human_detection_csv(csv, vid_dir)

    def __str__(self) -> str:
        return f"GroundTruth with {len(self.meteors)} detections, {self.n_unique_videos()} unique videos, and db dir:\n\t{self.vid_dir}"
    
    def __len__(self) -> int:
        return len(self.meteors)
    
    def __getitem__(self, key):
        return self.meteors[key]

    def n_unique_videos(self) -> int:
        return len(self.vids())

    def try_command(self, args: fmdt.args.Args) -> list[bool]:
        """Take a set of parameters defined by `args` and apply it to every meteor in this database.
        
        Return a list of booleans that are true if the corresponding meteor was
        detected. 
        """
        def try_comm(m: HumanDetection) -> bool:

            args.detect_args.vid_in_path = m.video_name

            print(f"Trying args on video {args.vid()}")
            
            res = args.detect()
            is_detected = is_meteor_detected(m, res.trk_list)
            if not is_detected:
                print(f"{res.args.detect_args.cmd()} {colored('unsuccessful', 'red')}")
            else:
                print(f"{res.args.detect_args.cmd()} {colored('successful', 'green')}")
            
            return is_detected

        # return [try_comm(m) for m in self.meteors if os.path.exists(m.video_name)]
        return [try_comm(m) for m in self.meteors]
    
    def vids(self) -> list[str]:
        """Return the list of unique videos that appear in this database"""
        return set([m.video_name for m in self.meteors])

    def test_span(self, light_min_min, light_min_max, diff):
        # Record the pairs that are successful, and the number of meteors detected 
        
        # n = (light_min_min - light_min_max) // diff
        print(f"test_span({light_min_min}, {light_min_max}, {diff})")

        min_max = []
        dets = []
        successes = []

        n = (light_min_max - light_min_min) / diff + 1

        # d_args = {
        #     "trk_out_path": "trk.txt",
        #     "trk_bb_path": "bb.txt",
        #     "timeout":  1
        # }

        args = fmdt.args.detect_args(timeout=0.5)

        for lmin in np.linspace(light_min_min, light_min_max, int(n)):

            # Get the list
            args.detect_args.light_min = lmin
            args.detect_args.light_max = lmin + diff 

            success_list = self.try_command(args)

            ndets = sum(success_list)

            min_max.append((lmin, lmin + diff)) 
            dets.append(ndets)
            successes.append(success_list)

        return min_max, dets, successes
    
    def draw_heatmap(self, lmin_min, lmax_max, n_intervals):

        successes = []
        diff = (lmax_max - lmin_min) / (n_intervals)

        min_max, _, successes = self.test_span(lmin_min, lmax_max - diff, diff)

        plot_gt(min_max, successes)


import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

def plot_gt(minmax: list[tuple[float, float]], successes: list[list[bool]]):

    lmin_min = minmax[0][0]
    lmin_max = minmax[-1][1]

    _, ax = plt.subplots()
    num_gt = len(successes[0])
    
    print(f"Num intervals: {len(minmax)}")
    print(f"Num ground truths: {num_gt}")

    y_ticks = []

    for i in range(len(minmax)):
        lmin = minmax[i][0]
        lmax = minmax[i][1]
        y_ticks.append(lmin)
        h = lmax - lmin
        w = 1
        print(f"lmin: {lmin}")

        # Now iterate horizontally
        for id in range(num_gt):
            if successes[i][id]:
                # print("Rect")
                ax.add_patch(Rectangle((id, lmin), w, h, facecolor = "green"))
            else:
                ax.add_patch(Rectangle((id, lmin), w, h, facecolor = "black"))

    y_ticks.append(lmin_max)

    plt.xlim([0, num_gt])
    plt.ylim([lmin_min, lmin_max])
    plt.yticks(y_ticks)
    plt.show()

# for d in range(len(MIN_MAX)):
#     plot_gt(MIN_MAX[d], SUCCESS[d])


# I want a function that automatically draws that heat map given lmin_min, lmin_max, and n




# Take a row in a csv file from https://docs.google.com/spreadsheets/d/1yYsMy0nnLAsqzTYOoEjof00wVu6oP91vCiLS7orxMpg/edit#gid=365172110
# and convert it to a HumanDetection object
def csv_to_human(csv_row: str) -> HumanDetection:
    entries = csv_row.split(',')
    return HumanDetection(entries[0],
                          int(entries[1]),
                          int(entries[2]),
                          float(entries[3]),
                          float(entries[4]),
                          float(entries[5]),
                          float(entries[6]))

# Take a single row of a pandas data frame of the csv data and return
# a python HumanDetection object
def df_row_to_human(row: pd.Series, dir_prepend: str = "") -> HumanDetection:

    # This function is performing NOOOO sanity checks
    # like do any of these fields even actually exist 
    # in our data frame??

    if dir_prepend[-1] != "/":
        dir_prepend = dir_prepend + "/"

    return HumanDetection(dir_prepend + row["video_name"],
                          int(row["start_frame"]),
                          int(row["end_frame"]),
                          float(row["start_x"]),
                          float(row["start_y"]),
                          float(row["end_x"]),
                          float(row["end_y"]))

def read_human_detection_csv(csv_filename: str, db_dir: str = "") -> list[HumanDetection]:
    df = pd.read_csv(csv_filename)
    return [df_row_to_human(df.iloc[i], db_dir) for i in range(len(df.index))]

def init_ground_truth(csv_filename: str = "human_detections.csv", vid_db_dir: str = "./") -> GroundTruth:
    return GroundTruth(csv=csv_filename, vid_dir=vid_db_dir)


# GROUND_TRUTH : list[HumanDetection] = None

# def init_ground_truth(database_file: str) -> list[HumanDetection]:
#     super(GROUND_TRUTH) = read_human_detection_csv(database_file)
#     return read_human_detection_csv(database_file)

# GROUND_TRUTH = read_human_detection_csv("human_detections.csv")

def is_meteor_detected(meteor: HumanDetection, tracking_list: list[TrackedObject]) -> bool:

    if tracking_list is None:
        return False

    # keep only meteors
    detected_m = [m for m in tracking_list if m.is_meteor()]

    for tracked in detected_m: 
        if are_objects_the_same(meteor, tracked):
            return True

    return False 

def are_objects_the_same(meteor: HumanDetection, tracked_obj: TrackedObject, log: bool = False) -> bool:
    """Check if a HumanDetected meteor is the same as a tracked obj
    
    A tracked object is a dictionary with keys:
        start_frame,
        end_frame,
        start_x,
        end_x,
        start_y,
        end_y 
    """

    # Are the lifetimes the same???
    # ====================== Lifetime ===========================
    lifetime_meteor = (meteor.start_frame, meteor.end_frame)
    lifetime_object = (tracked_obj.start_frame, tracked_obj.end_frame)

    # meteor:                        [***********************]           
    # object:        [------]
    if lifetime_object[1] < lifetime_meteor[0]:
        return False
    
    # meteor:                       [***********************]           
    # object:                                                       [------]
    if lifetime_object[0] > lifetime_meteor[1]:
        return False
    
    # =============== Compare Direction ==========================
    angle_meteor_rad = meteor.direction() 
    angle_object_rad = tracked_obj.direction() 

    MAX_ANGLE_DIFF = 0.5 #!!! ARBITRARY

    if log:
        print(f"Meteor has angle: {angle_meteor_rad}")
        print(f"Object has angle: {angle_object_rad}")

    if abs(angle_meteor_rad - angle_object_rad) > MAX_ANGLE_DIFF:
        if log:
            print("Angles are too far apart")
        return False
    

    # ================ Compare Flight paths ==========================
    # At this point, there are only four cases left
    #   TRUTH      [***********************]           
    # case 1: 
    #         [------]
    #
    # case 2:
    #                                 [----------]
    #
    # case 3:
    #                [---------------]
    #
    # cse 4:
    #       [---------------------------------------]


    # We'll start with case 1 and 2
    if tracked_obj.start_frame < meteor.start_frame: # CASE 1!!!
        # Let's ALSO create a flight path for our object, starting at frame meteor.start_frame
        frames_object = [i for i in range(lifetime_meteor[0], lifetime_object[1] + 1)] 

    elif tracked_obj.end_frame > meteor.end_frame: # Case 2!!!!

        frames_object = [i for i in range(lifetime_object[0], lifetime_meteor[1] + 1)]

    elif tracked_obj.end_frame > meteor.end_frame and tracked_obj.start_frame < meteor.start_frame: # Case 4!!!!

        frames_object = [i for i in range(lifetime_meteor[0], lifetime_meteor[1] + 1)]

    else: # Case 3
        frames_object = [i for i in range(lifetime_object[0], lifetime_object[1] + 1)]

    # Now we want to compare the two flight paths!
    # TODO choisir judiciement un epsilon
    MAX_DIST = 10 # ????????????????????????????????????????????????? 

    # If the any two points in the flight path are too far away, then they are not the 
    # same meteor.
    for f in frames_object:

        pos_meteor = meteor.interpolate_pos(f) 
        pos_object = tracked_obj.interpolate_pos(f - 1)

        # print(f"f: {f}, pos_m: {meteor.interpolate_pos(f)}, pos_o: {tracked_obj.interpolate_pos(f)}")

        # print(pos_meteor)
        # print(po)

        sqr = lambda x : x * x

        dist = math.sqrt(sqr(pos_object[0] - pos_meteor[0]) + sqr(pos_object[1] - pos_meteor[1]))

        if dist > MAX_DIST:
            # print(f"Position at frame {f} is too far apart (dist: {dist})")
            return False
        
    return True
        

def vary_light_intervals(
        vid: str,
        truth: HumanDetection,
        offset: int,
        light_min_start: int,
        light_min_end: int,
        k_trials: int,
        log: bool = False
    ) -> list[fmdt.args.Args]:
    """Vary light min and light max to try and detect the truth
    
    Return a list of args that succesfully detected this truth
    """

    args = []

    for lmin in np.linspace(light_min_start, light_min_end, k_trials, dtype=int):


        light_min = int(lmin)
        light_max = light_min + offset

        track_file = "tracks.txt"
        track_bb_path = "track_bb.txt"

        vname, ext = fmdt.utils.decompose_video_filename(vid)
        vname = f"{vname}_off{offset}"

        res = fmdt.detect(vid, 
                        trk_out_path=track_file,
                        log=True,
                        light_min=light_min,
                        light_max=light_max,
                        trk_all=True,
                        trk_bb_path=track_bb_path)

        if len(res.tracking_list) != 0:

            if not os.path.isdir(vname):
                os.mkdir(vname)

            if truth.is_detected_in_list(res.tracking_list):
                addon = "_detected"
                if truth.is_detected_in_list([t for t in res.tracking_list if t.is_meteor()]):
                    addon = f"{addon}_as_meteor"
                    args.append(res)
            else:
                addon = ""

            fmdt.visu(vid, track_file, track_bb_path, f"{vname}/lmin{light_min}_lmax{light_max}{addon}.{ext}", show_id=True)

        if log: 
            meteors = [m for m in res.tracking_list if m.is_meteor()]
            for m in meteors:
                print(m)

    return args

def main() -> None:
    # dets = read_human_detection_csv("human_detection.csv")
    # for d in dets:
    #     print(d)

    m = HumanDetection("demo.mp4", 180, 184, 797.2, 459.7, 810.7, 482.0)

    tracking_list = fmdt.core.extract_all_information("demo.txt")
    meteors = [obj for obj in tracking_list if obj.is_meteor()]

    # print(m)
    # print(tracking_list)

    # for obj in tracking_list:
        # print(are_objects_the_same(m, obj))

    obj_to_hum = lambda obj, name: HumanDetection(name, obj.start_frame, obj.end_frame, obj.start_x, obj.start_y, obj.end_x, obj.end_y)

    hums = [obj_to_hum(obj, "demo.mp4") for obj in meteors]

    # for i in range(len(hums)):
    # for i in range(5):
        # print(meteors[i])
        # print(hums[i])

    # for h in hums:
        # print(h)

    # Now let's test to see if each "human" detection was found in the tracking list!

    print(m)
    print(f"Is m detected: {is_meteor_detected(m, tracking_list)}")

    for h in hums:
        print(f"Is h: {h} detected? {is_meteor_detected(h, tracking_list)}")


    # for h in hums:
        # print(h)

    # for obj in meteors:
        # print(are_objects_the_same(hums[15], obj))

    print(len(hums))

    # print(GROUND_TRUTH)
    


"""Module dealing with downloading files and stuff"""
import os
import requests
from appdirs import user_data_dir 

# Let's define some functions that will fetch our data base files

__DATA_DIR = user_data_dir("fmdt_python")

__GITHUB_PREFIX = "https://raw.githubusercontent.com/ejovo13/fmdt_python_clean/build"

__DRACO6_CSV_URL  = __GITHUB_PREFIX + "/human_detections_draco6.csv"
__DRACO12_CSV_URL = __GITHUB_PREFIX + "/human_detections_draco12.csv"
__DRACO_CSV_URL = __GITHUB_PREFIX + "/human_detections.csv"

# "name_of_local_file": "url_of_db"
csv_dict = {
    "human_detections_draco6.csv": __DRACO6_CSV_URL,
    "human_detections_draco12.csv": __DRACO12_CSV_URL,
    "human_detections.csv": __DRACO_CSV_URL
}


def download_file(
        filename: str,
        url: str,
        dir: str = "./",
        log = False,
        overwrite = False
    ) -> bool:
    """Download files whose contents are stored in text"""
    if not dir is None:
        filename = dir + "/" + filename

    if os.path.exists(filename) and not overwrite:
        if (log):
            print(f"{filename} already exists, not downloading anything")
        return False
    else:
        if (log):
            print(f"Downloading csv file from {url}")
        r = requests.get(url)
        with open(filename, "w") as csv: 
            csv.write(r.text)

        return True
    
def download_binary_file(
        filename: str,
        url: str,
        dir: str = "./",
        log = False,
        overwrite = False
    ) -> bool:

    if not dir is None:
        filename = dir + "/" + filename

    if os.path.exists(filename) and not overwrite:
        if (log):
            print(f"{filename} already exists, not downloading anything")
        return False
    else:
        if (log):
            print(f"Downloading csv file from {url}")

        r = requests.get(url, stream=True)
        with open(filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024): 
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)

        return True
    
def download_draco6_csv(filename: str = "human_detections_draco6.csv", dir = __DATA_DIR) -> bool:
    """Return True if downloading"""
    return download_file(filename, __DRACO6_CSV_URL, dir=dir)


def download_draco12_csv(filename: str = "human_detections_draco12.csv", dir = __DATA_DIR) -> bool:
    return download_file(filename, __DRACO12_CSV_URL, dir=dir)

def download_draco_csv(filename: str = "human_detections.csv", dir = __DATA_DIR) -> bool:
    return download_file(filename, __DRACO_CSV_URL, dir=dir)

def download_csvs(log = False, overwrite = False, dir = __DATA_DIR) -> None:
    for (file, url) in csv_dict.items():
        download_file(file, url, dir, log, overwrite)


def download_videos_db(filename = "videos.db", log = False, overwrite = False, dir = __DATA_DIR):
    """Download videos.db"""
    url  = "https://github.com/ejovo13/fmdt_python_clean/raw/build/videos.db"

    download_binary_file(filename, url, dir, log, overwrite)

def download_dbs(overwrite = True):
    download_videos_db(overwrite=overwrite)

def download_demo_mp4(filename: str = "demo.mp4"):

    url = "https://lip6.fr/adrien.cassagne/data/tauh/in/2022_05_31_tauh_34_meteors.mp4"

    if os.path.exists(filename):
        print(f"{filename} already exists, not downloading")
        return False
    else:
        print(f"Downloading mp4 file from {url}...")
        r = requests.get(url)
        with open(filename, "wb") as file:
            for chunk in r.iter_content(chunk_size=255): 
                if chunk: # filter out keep-alive new chunks
                    file.write(chunk)

            print("Done")



"""Companion script for https://fmdt-python-clean.readthedocs.io/en/latest/howto/2_test_db/"""
import fmdt
import os
import random

vid_dir = "/home/ejovo/Videos/Watec6mm"        # Path to draco6 videos on my machine
draco6_db = "human_detections_draco6.csv"

if not os.path.exists(vid_dir):
    print("Please specify the database directory")

assert os.path.exists(vid_dir), f"vid_dir: '{vid_dir}' not a path"
assert os.path.exists(draco6_db), f"draco6_db: '{draco6_db}' does not exist"

gt6 = fmdt.GroundTruth(csv=draco6_db, vid_dir=vid_dir)

d_args = {
    "light_min": 250,
    "light_max": 253,
    "trk_out_path": "trk.txt",
    "trk_bb_path": "bb.txt",
    "timeout": 1          # timeout in seconds for the fmdt-detect subprocess
}

args = fmdt.Args(detect_args=d_args)

args.detect_args["light_min"] = 240
args.detect_args["light_max"] = 250

# success_list = gt6.try_command(args)

# print(success_list)
# print(f"GroundTruths detected: {sum(success_list)}")

# dets = [gt6.meteors[i] for i in range(len(gt6.meteors)) if success_list[i]]
# ids  = [i for i in range(len(gt6.meteors)) if success_list[i]]


# Sample randomly in the range 

LIGHT_MIN_MIN = 230
LIGHT_MIN_MAX = 240
DIFF_MAX = 30

k_trials = 7 

# with open("gt_ids_det.csv", "a") as file:

#     for k in range(k_trials):

#         light_min  = random.randint(LIGHT_MIN_MIN, LIGHT_MIN_MAX)
#         light_diff = random.randint(1, min(255 - light_min, DIFF_MAX))

#         light_max = light_min + light_diff

#         args.detect_args["light_min"] = light_min 
#         args.detect_args["light_max"] = light_max 

#         success_list = gt6.try_command(args)

#         dets = [gt6.meteors[i] for i in range(len(gt6.meteors)) if success_list[i]]
#         ids  = [i for i in range(len(gt6.meteors)) if success_list[i]]

#         for i in ids:
#             file.write(str(i) + ",")


# Let's try to hold the interval constant, and systematically check all of the light_min values

DIFF = 5
LIGHT_MIN_MIN = 245
LIGHT_MIN_MAX = 255 - DIFF 

import numpy as np

# for k in range(k_trials):
def test_span_gt6(light_min_min, light_min_max, diff):
    # Record the pairs that are successful, and the number of meteors detected 

    min_max = []
    dets = []


    for lmin in np.linspace(light_min_min, light_min_max, 3):

        # Get the list
        args.detect_args["light_min"] = lmin
        args.detect_args["light_max"] = lmin + diff 

        success_list = gt6.try_command(args)

        ndets = sum(success_list)

        min_max.append((lmin, lmin + diff)) 
        dets.append(ndets)

    return min_max, dets

MIN_MAX = []
DETS = []

# for diff in [1, 3, 5]:
#     min_max, dets = test_span_gt6(LIGHT_MIN_MIN, LIGHT_MIN_MAX, diff)
#     MIN_MAX.append(min_max)
#     DETS.append(dets)

print(MIN_MAX)
print(DETS)

# for d in dets:
#     print(f"Detected: {d}")

# for i in ids:
#     print(i)

# We can work with a subset of the ground truth data sets if we only use

# visible = [0, 2, 3, 6, 34, 35]
visible = [0, 2, 3, 5, 6, 24, 28, 34, 35]
red_mets = [gt6.meteors[i] for i in visible]

gt_small = fmdt.GroundTruth("human_detections_draco6.csv")
gt_small.meteors = red_mets

print(len(gt_small.meteors))

def test_span_gt_small(light_min_min, light_min_max, diff):
    # Record the pairs that are successful, and the number of meteors detected 

    min_max = []
    dets = []
    successes = []


    for lmin in np.linspace(light_min_min, light_min_max, 2):

        # Get the list
        args.detect_args["light_min"] = lmin
        args.detect_args["light_max"] = lmin + diff 

        success_list = gt_small.try_command(args)

        ndets = sum(success_list)

        min_max.append((lmin, lmin + diff)) 
        dets.append(ndets)
        successes.append(success_list)

    return min_max, dets, successes

MIN_MAX = []
DETS = []
SUCCESS = []

# for diff in [1, 3, 5]:
# for diff in [5]:
#     min_max, dets, success = test_span_gt_small(LIGHT_MIN_MIN, LIGHT_MIN_MAX, diff)
#     MIN_MAX.append(min_max)
#     DETS.append(dets)
#     SUCCESS.append(success)

# print(MIN_MAX)
# print(DETS)
# print(SUCCESS)

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

def plot_gt(minmax: list[tuple[float, float]], successes: list[list[bool]]):

    lmin_min = minmax[0][0]
    lmin_max = minmax[-1][1]

    _, ax = plt.subplots()
    num_gt = len(successes[0])
    
    print(f"Num intervals: {len(minmax)}")
    print(f"Num ground truths: {num_gt}")

    y_ticks = []

    for i in range(len(minmax)):
        lmin = minmax[i][0]
        lmax = minmax[i][1]
        y_ticks.append(lmin)
        h = lmax - lmin
        w = 1
        print(f"lmin: {lmin}")

        # Now iterate horizontally
        for id in range(num_gt):
            if successes[i][id]:
                # print("Rect")
                ax.add_patch(Rectangle((id, lmin), w, h, facecolor = "green"))
            else:
                ax.add_patch(Rectangle((id, lmin), w, h, facecolor = "black"))

    y_ticks.append(lmin_max)

    plt.xlim([0, num_gt])
    plt.ylim([lmin_min, lmin_max])
    plt.yticks(y_ticks)
    plt.show()

# for d in range(len(MIN_MAX)):
#     plot_gt(MIN_MAX[d], SUCCESS[d])


# I want a function that automatically draws that heat map given lmin_min, lmin_max, and n

def draw_heatmap(gt: fmdt.GroundTruth, lmin_min, lmax_max, n_intervals):

    min_max = []
    dets = []
    successes = []
    diff = (lmax_max - lmin_min) / (n_intervals)

    for lmin in np.linspace(lmin_min, lmax_max - diff, n_intervals):

        # Get the list
        args.detect_args["light_min"] = lmin
        args.detect_args["light_max"] = lmin + diff 

        success_list = gt.try_command(args)

        ndets = sum(success_list)

        min_max.append((lmin, lmin + diff)) 
        dets.append(ndets)
        successes.append(success_list)

    plot_gt(min_max, successes)


# draw_heatmap(gt_small, 150, 255, 2)


# draw_heatmap(gt_small, 240, 255, 15)

# draw_heatmap(gt6, 240, 255, 10)

draco12 = "human_detections_draco12.csv"
vid_dir12 = "/home/ejovo/Videos/Watec12mm"

gt12 = fmdt.GroundTruth(draco12, vid_dir12)

draw_heatmap(gt12, 150, 255, 155)


# Convert the results obtained from `run_args.sh` into a three arrays of the same size:
#
# light_min
# light_max
# n_detected

# file = "draco6_random_lminmax_results.txt"
# file = "data_all_k30j10.txt"
file = "data_all.txt"
# file = "draco12_k30_j10.txt"

# Convert a '[1, 2, 3]' to [1, 2, 3]
def strlist_to_listint(strlist: str) -> list[int]:
    red = strlist[1:-1]
    ints = red.split(',')
    return [int(i.strip()) for i in ints]

print(strlist_to_listint('[1, 2, 3]'))

with open(file) as f:
    lines = [strlist_to_listint(l.strip()) for l in f.readlines()]

    lmins = lines[0::3]
    lmaxs = lines[1::3]
    dets  = lines[2::3]

    lmins = [j for i in lmins for j in i]    
    lmaxs = [j for i in lmaxs for j in i]    
    dets  = [j for i in dets  for j in i]    

    print(lmins)
    print(lmaxs)
    print(dets)


"""Test out the split window videos"""

import fmdt
import os

dir = "/home/ejovo/Videos/Window/window_3_sony_0400-0405UTC"

files = os.listdir(dir)

print(files)

for f in files:
    full_path = dir + "/" + f
    
    (fmdt
        .detect(
            vid_in_path=full_path,
            trk_out_path="trk.txt",
            trk_bb_path="bb.txt",
            light_min=75,
            light_max=100)
        .visu())

# Let's play around with the second file

v = files[1]




# Start implementing the functions needed to check fmdt

import fmdt
import fmdt.truth

og = "2022_05_31_tauh_34_meteors.mp4"

args = fmdt.detect_args(trk_out_path="trk.txt")
meteors = fmdt.truth.load_meteors_file("meteors.txt", "2022_05_31_tauh_34_meteors.mp4")

print(meteors)

v = fmdt.Video(og, fmdt.VideoType.WINDOW)
v.evaluate_args(args, meteors)

print(v)
print(v.full_path())
print(v.exists())

v.detect()

print(fmdt.load_config())


"""Learn how to play around with sql lite"""

import os
import fmdt
import fmdt.config

con = fmdt.config.load_config()
print(con)

files = os.listdir(con.d6)

all_vids = fmdt.config.get_local_vids()

print(all_vids)
print(len(all_vids))


"""
Core functions used in the processing of output for fmdt: https://github.com/alsoc/fmdt

Public API:
    extract_key_information
    extract_all_information
    split_video_at_meteors
"""
import os
import fmdt.utils as utils
from enum import Enum
import math
from termcolor import colored
import re

red = lambda s: colored(s, "red")
black = lambda s: colored(s, "black")
blue = lambda s: colored(s, "blue")
green = lambda s: colored(s, "green")

# Structure of tracking output table of fmdt-detect after 
# each line gets stripped using whitespace as a delimiter
class TrackingTable:

    OBJECT_ID   = 0
    START_FRAME = 2
    START_X     = 4
    START_Y     = 6
    END_FRAME   = 8
    END_X       = 10
    END_Y       = 12
    OBJECT_TYPE = 14


class ObjectType(Enum):
    """Simple Enum to distinguish between meteor, star and noise TrackedObjects"""
    METEOR = 0
    STAR   = 1
    NOISE  = 2

    @staticmethod
    def from_str(s: str):
        slow = s.lower()
        if slow == "meteor":
            return ObjectType.METEOR
        elif slow == "star":
            return ObjectType.STAR
        else:
            return ObjectType.NOISE


    
class TrackedObject:
    """A `TrackedObject` is used to store the celestial objects detected by `fmdt-detect`

    If we call `fmdt-detect` and store the track results to a file with:

    >>> fmdt.detect("demo.mp4", trk_out_path="tracks.txt")  

    Then we can load in the detected objects as a list of `TrackedObject`:

    >>> objects = fmdt.read_tracks_file("tracks.txt")

    We are effectively loading in the tracking table:

    # -------||---------------------------||---------------------------||---------
    #  Track ||           Begin           ||            End            ||  Object
    # -------||---------------------------||---------------------------||---------
    # -------||---------|--------|--------||---------|--------|--------||---------
    #     Id || Frame # |      x |      y || Frame # |      x |      y ||    Type
    # -------||---------|--------|--------||---------|--------|--------||---------
       {tid} ||  {fbeg} | {xbeg} | {ybeg} ||  {fend} | {xend} | {yend} || {otype}
    """
    def __init__(
        self,
        id: int,
        start_frame: int,
        start_x: float,
        start_y: float,
        end_frame: int,
        end_x: float,
        end_y: float,
        type: ObjectType
        ): 

        self.id = id
        self.start_frame = start_frame
        self.start_x = start_x
        self.start_y = start_y
        self.end_frame = end_frame
        self.end_x = end_x
        self.end_y = end_y
        self.type = type
        
    def is_meteor(self) -> bool:
        return self.type == ObjectType.METEOR

    def is_star(self) -> bool:
        return self.type == ObjectType.STAR

    def is_noise(self) -> bool:
        return self.type == ObjectType.NOISE
    
    def type_str(self) -> str:
        """Return the ObjectType of this object as a string"""
        if self.is_meteor():
            return "Meteor" 
        elif self.is_star():
            return "Star" 
        else:
            return "Noise" 
    
    def __str__(self) -> str:

        return f"<{self.type_str()} ({self.start_frame}, {self.end_frame}), pos0: ({self.start_x, self.start_y}), posT: ({self.end_x}, {self.end_y})>"
    
    def __repr__(self) -> str:
        return f"<{self.type_str()} ({self.start_frame}, {self.end_frame})>"
    

    def delta_x(self) -> float:
        """Displacement in the x_axis direction"""
        return self.end_x - self.start_x
    
    def delta_y(self) -> float:
        """Displacement in the y direction"""
        return self.end_y - self.start_y
    
    def displacement(self) -> tuple[float, float]:
        return (self.delta_x(), self.delta_y())
    
    def lifetime(self) -> tuple[int, int]:
        return (self.start_frame, self.end_frame)
    
    def nframes_alive(self) -> int:
        return self.end_frame - self.start_frame

    def slope(self) -> float:
        if self.delta_x() == 0:
            return float("inf")
        return self.delta_y() / self.delta_x()
    
    def direction(self) -> float:
        """Return the angle of the displacement vector of this meteor. Units are radians"""
        return math.atan2(self.delta_y(), self.delta_x())

    def dx(self) -> float:
        return self.delta_x() / self.nframes_alive()
    
    def interpolate_pos(self, frame_n: int) -> tuple[float, float]:
        f_prime = frame_n - self.start_frame
        xi = self.start_x + (self.dx() * f_prime)
        yi = self.start_y + (self.dx() * self.slope()) * f_prime
        return (xi, yi)


def read_tracks_file(tracks_filename: str) -> list[TrackedObject]:
    return extract_all_information(tracks_filename)


def extract_key_information(detect_tracks_in: str) -> list[dict]:
    """Extract key information from a detect_tracks.txt file.

    "Key" information refers to the start frame, end frame, and type of 
    object detected

    Parameters
    ----------
    `detect_tracks_in` (str): The name of a file whose content is the output
        of fmdt_detect

    Returns
    -------
    dict_array (list[dict]): A list of dictionaries of the form
        { 
            "type": <"meteor" | "noise" | "start">
            "frame_start": <int>
            "frame_end": <int>
        }
        where each item in the list corresponds to a single object detected by 
        `fmdt-detect`
    """

    # Utility function to convert a line of interest to a dictionary
    def line_to_dict(split_line: list[str]):

        return {
            "type":         split_line[TrackingTable.OBJECT_TYPE],
            "start_frame":  int(split_line[TrackingTable.START_FRAME]),
            "end_frame":    int(split_line[TrackingTable.END_FRAME])
        }

    # Utility boolean function to extract only the important lines
    interesting_line = lambda line: (
        (" meteor" in line) or (" star" in line) or (" noise" in line)
    )
    
    # Processing of the actual file
    file_tracks = open(detect_tracks_in)
    file_lines  = file_tracks.readlines()
    dict_array = []

    for line in file_lines:
        if interesting_line(line):
            dict_array.append(line_to_dict(line.split()))

    return dict_array

def nframes_processed(detect_tracks_in: str) -> int:

    if not os.path.exists(detect_tracks_in):
        return 0

    with open(detect_tracks_in) as file:

        lines = [l for l in file.readlines() if "-> Processed frames =" in l]
        if len(lines) == 0:
            return 0
        line_split = lines[0].split()
        out = int(line_split[-1])

    return out


def extract_all_information(detect_tracks_in: str) -> list[TrackedObject]:
    """Extract all tracking information from a detect_tracks.txt file.

    Parameters
    ----------
    `detect_tracks_in` (str): The name of a file whose content is the output
        of fmdt_detect

    Returns
    -------
    list_objects (list[TrackedObject]): A list of TrackedObject
        { 
            "id":           <int>,
            "start_frame":  <int>,
            "start_x":      <float>,
            "start_y":      <float>,
            "end_frame":    <int>,
            "end_x":        <float>,
            "end_y":        <float>,
            "type":         <"meteor" | "noise" | "start">,
        }
        where each item in the list corresponds to a single object detected by 
        `fmdt-detect`
    """

    # Utility function to convert a line of interest to a dictionary
    # def line_to_dict(split_line: list[str]):

    #     return {
    #         "id":           int(split_line[TrackingTable.OBJECT_ID]),
    #         "start_frame":  int(split_line[TrackingTable.START_FRAME]),
    #         "start_x":      float(split_line[TrackingTable.START_X]),
    #         "start_y":      float(split_line[TrackingTable.START_Y]),
    #         "end_frame":    int(split_line[TrackingTable.END_FRAME]),
    #         "end_x":        float(split_line[TrackingTable.END_X]),
    #         "end_y":        float(split_line[TrackingTable.END_Y]),
    #         "type":         split_line[TrackingTable.OBJECT_TYPE]
    #     }

    if not os.path.exists(detect_tracks_in):
        return []

    def line_to_obj(split_line: list[str]) -> TrackedObject:
        return TrackedObject(int  (split_line[TrackingTable.OBJECT_ID]),
                             int  (split_line[TrackingTable.START_FRAME]),
                             float(split_line[TrackingTable.START_X]),
                             float(split_line[TrackingTable.START_Y]),
                             int  (split_line[TrackingTable.END_FRAME]),
                             float(split_line[TrackingTable.END_X]),
                             float(split_line[TrackingTable.END_Y]),
                             ObjectType.from_str(split_line[TrackingTable.OBJECT_TYPE]))
      
    # Utility boolean function to extract only the important lines
    interesting_line = lambda line: (
        (" meteor" in line) or (" star" in line) or (" noise" in line)
    )
    
    # Processing of the actual file
    file_tracks = open(detect_tracks_in)
    file_lines  = file_tracks.readlines()
    file_tracks.close()

    return [line_to_obj(line.split()) for line in file_lines if interesting_line(line)]


def split_video_at_meteors(
        video_filename: str,
        detect_tracks_in: str,
        nframes_before=3,
        nframes_after=3,
        overwrite=False,
        exact_split: bool = False,
        log: bool = False
    ) -> None:
    """
    Split a video into small segments of length (nframes_before + nframes_after + sequence_length) frames
    for each meteor detected 

    Parameters
    ----------
    `video_filename` (str): Filename of video to split 
    `detect_tracks_in` (str): Filename of tracks recorded by a call to 
        fmdt.detect(video_filename, trk_out_path=detect_tracks_in)
    `nframes_before` (int): Number of frames to extract before the meteor sequence begins
        (Default value of 3)
    `nframes_after` (int): Number of frames to extract after the meteor sequence ends
        (Default value of 3)
    `overwrite` (bool): Tells ffmpeg to overwrite (True) the generated output videos if they already
        exist. If false then ffmpeg will ask for manual confirmation
        (Default value of False)
    `exact_split` (bool): Determine whether to extract the _exact_ frames requested (True) - which requires loading
        the entire video into memory as a numpy array - or approximate the frames requested using ffmpeg's 
        seeking functionality (False). For long videos with high resolution, use False otherwise the program 
        might crash. 
        (Default value of False)
    `log` (bool): When True, print to console the current action being performed
    """
    utils.assert_file_exists(detect_tracks_in)

    # Preprocessing of information held in `detect_tracks_in`
    tracking_list = extract_key_information(detect_tracks_in)
    tracking_list = utils.retain_meteors(tracking_list)
    seqs = utils.separate_meteor_sequences(tracking_list)

    split_video_at_intervals(video_filename, seqs, nframes_before, nframes_after, overwrite, exact_split, log, condense=False)

def get_dir(filename: str):
    return '/'.join(filename.split()[:-1])

def split_video_at_intervals(
        video_filename: str,
        start_end: list[tuple[int, int]],
        nframes_before=3,
        nframes_after=3,
        overwrite=False,
        exact_split: bool = False,
        log: bool = False,
        condense: bool = True
    ) -> None:
    """
    Split a video into small segments of length (nframes_before + nframes_after + sequence_length) frames
    for each meteor detected 

    Parameters
    ----------
    `video_filename` (str): Filename of video to split 
    `start_end` (list[tuple[int, int]]): list of (frame_start, frame_end) pairs to split this video at
    `nframes_before` (int): Number of frames to extract before the meteor sequence begins
        (Default value of 3)
    `nframes_after` (int): Number of frames to extract after the meteor sequence ends
        (Default value of 3)
    `overwrite` (bool): Tells ffmpeg to overwrite (True) the generated output videos if they already
        exist. If false then ffmpeg will ask for manual confirmation
        (Default value of False)
    `exact_split` (bool): Determine whether to extract the _exact_ frames requested (True) - which requires loading
        the entire video into memory as a numpy array - or approximate the frames requested using ffmpeg's 
        seeking functionality (False). For long videos with high resolution, use False otherwise the program 
        might crash. 
        (Default value of False)
    `log` (bool): When True, print to console the current action being performed

    Examples
    --------

    Create two segments from frames [~10, ~20] and [~100, ~200]    

    >>> fmdt.split_video_at_intervals("demo.mp4", start_end=[(10, 20), (100, 200)])

    Create two segments split exactly at [10, 20] and [100, 200] 

    >>> fmdt.split_video_at_intervals(
            "demo.mp4",
            start_end=[(10, 20), (100, 200)],
            exact_split=True,
            nframes_before = 0,
            nframes_after = 0
        )


    """
    utils.assert_file_exists(video_filename)

    # Preprocessing of information held in `detect_tracks_in`

    if condense:
        seqs = utils.condense_start_end(start_end)
    else:
        seqs = start_end
    video_name, extension = utils.decompose_video_filename(video_filename) 

    # Bookkeeping for formatting the name of the output videos
    max_digits = len(str(seqs[-1][1]))
    format_str = '0' + str(max_digits)

    # Selection of splitting algorithm
    if exact_split:
        
        # Querying of video information, extraction of frames
        frames = utils.convert_video_to_ndarray(video_filename, log=log)
        frame_rate = utils.get_avg_frame_rate(video_filename)
        total_frames, _, _, _ = frames.shape
        seq_video_name = lambda seq: (
            f'{video_name}/f{format(seq[0], format_str)}-{format(seq[1], format_str)}.{extension}'
        )

        def exact_splitting(seq) -> None:
            """Function call that will convert a 'meteor seq' to a frame perfect video sequence"""

            # Ensure that f_start and f_end are valid
            f_start = s[0] - nframes_before if s[0] - nframes_before >= 0 else 0
            f_end   = s[1] + nframes_after  if s[1] + nframes_after  <= total_frames else total_frames
            frames_seq = frames[f_start:f_end, :, :, :]
            utils.convert_ndarray_to_video(seq_video_name(s), frames_seq, frame_rate, log=log)

        splitting_algorithm = exact_splitting
    
    else:

        seq_video_name = lambda seq: (
            f'{video_name}/f{format(seq[0], format_str)}-{format(seq[1], format_str)}_.{extension}'
        )

        def approx_splitting(seq) -> None:
            """Function call that will convert a 'meteor seq' to an approximate video sequence"""

            f_start = s[0] - nframes_before if s[0] - nframes_before >= 0 else 0
            f_end   = s[1] + nframes_after
            utils.extract_video_frames(video_filename, f_start, f_end,
                                       seq_video_name(s), overwrite=overwrite)

        splitting_algorithm = approx_splitting

    for s in seqs:
        splitting_algorithm(s)


"""Store the results of a run to fmdt-detect"""
import os
import re
import pandas as pd
import fmdt.args
import fmdt.core

def retrieve_all_nroi(log_path: str) -> list[int]:
    """
    Assumes the format: '# Frame n°00247 (t) -- Regions of interest (RoI) [64]: \n'
    """
    frames = os.listdir(log_path)
    rxp = r'\[(\d*)\]'

    rxp_comp = re.compile(rxp)

    def contains_roi(line: str) -> bool:
        return "(RoI)" in line

    def retrieve_single_nroi(frame: str):
        with open(log_path + "/" + frame, "r") as file:
            lines = file.readlines()
            rois = [l for l in lines if contains_roi(l) and "(t)" in l]
            tokens = rois[0].split() 
            n_roi = [n for n in tokens if rxp_comp.search(n)]
            return int(n_roi[0][1:-2])

    return [retrieve_single_nroi(f) for f in frames]

def retrieve_all_nassociations(log_path: str) -> list[int]:

    frames = os.listdir(log_path)[1:]

    rxp = r'\[\d*\]'

    rxp_comp = re.compile(rxp)

    def retrieve_single_nassociations(frame: str):
        with open(log_path + "/" + frame, "r") as file:
            lines = file.readlines()
            assocs = [l for l in lines if "Associations" in l]
            tokens = assocs[0].split() 
            n_roi = [n for n in tokens if rxp_comp.search(n)]
            return int(n_roi[0][1:-2])

    return [retrieve_single_nassociations(f) for f in frames]

def retrieve_mean_err_std_dev(filename: str) -> tuple[float, float]:
    """This function is JANK. It only depends on the fact that in a log file the only things that have 19
    columns are the rows containing the mean_err and std_dev!"""
    with open(filename) as file:
        lines = file.readlines()
        for l in lines:
            l_split = l.split()
            if len(l_split) == 19:
                mean_err = float(l_split[-3])
                std_dev  = float(l_split[-1])
                return (mean_err, std_dev)
            
def retrieve_all_mean_errs(log_path: str) -> list[float]:

    frames = os.listdir(log_path)[1:]

    def mean_err(filename) -> float:
        m, _ = retrieve_mean_err_std_dev(log_path + "/" + filename)
        return m
    
    return [mean_err(f) for f in frames]

def retrieve_all_std_devs(log_path: str) -> list[float]:

    frames = os.listdir(log_path)[1:]

    def mean_err(filename) -> float:
        _, s = retrieve_mean_err_std_dev(log_path + "/" + filename)
        return s
    
    return [mean_err(f) for f in frames]

def retrieve_log_info(log_path: str) -> tuple[list[int], list[int], list[float], list[float]]:
    """Return [nrois], [nassocs], [mean_errs], [std_devs]"""

    print(f"Trying to retrieve log info here: {log_path}")

    nrois = retrieve_all_nroi(log_path)
    nassocs = retrieve_all_nassociations(log_path)
    mean_errs = retrieve_all_mean_errs(log_path)
    std_devs = retrieve_all_std_devs(log_path)

    return nrois, nassocs, mean_errs, std_devs


def retrieve_log_df(log_path: str) -> pd.DataFrame:

    assert os.path.exists(log_path), f"log_path {log_path} does not exist"

    nroi, nassoc, mean_err, std_dev = retrieve_log_info(log_path)

    return pd.DataFrame({
        "nroi": nroi,
        "nassoc": [0] + nassoc,
        "mean_err": [0.0] + mean_err,
        "std_dev": [0.0] + std_dev
    })

class DetectionResult:

    def __init__(self, nframes: int, df: pd.DataFrame, args: fmdt.args.Args, trk_list: list[fmdt.core.TrackedObject], video = None):

        self.nframes = nframes
        self.df = df
        self.args = args 
        self.trk_list = trk_list
        self.video = video

    def detect(self):
        return self.args.detect()

    def visu(self):
        return self.args.visu()

    def n_meteors_detected(self) -> int:
        return len([m for m in self.trk_list if m.is_meteor()])

    def n_stars_detected(self) -> int:
        return len([s for s in self.trk_list if s.is_star()])

    def n_noise_detected(self) -> int:
        return len([n for n in self.trk_list if n.is_noise()])

    def trk_list_summary(self) -> str:
        return f"objects in trk_list: {self.n_meteors_detected()} meteor(s), {self.n_stars_detected()} star(s), {self.n_noise_detected()} noise"
    
    def __str__(self) -> str:
        a = f"fmdt.res.DetectionResult with args digest: {self.args.detect_args.digest()[0:16]}"
        b = f"\n{self.trk_list_summary()}"
        
        c = ""
        if not self.df is None:
            c = f"\n{str(self.df)}"

        return a + b + c
    
    def check(self, gt_path: str = None, stdout: str = "check.txt", log = False):
        """Call fmdt-check with these results
        
        Parameters
        ----------
        gt_path (str): Full path to ground truth file. Used when dealing with a call to fmdt.detect directly and not
            interfacing with a Video object. When calling with the function chain Video.detect().check(), we don't need
            to pass in the gt_path
        
        """

        assert not self.args.detect_args.trk_out_path is None, "To call fmdt.check we must specify the `trk_out_path`"

        if gt_path is None:
            # If we dont have ground truth meteors, assure that the meteor acutally exists
            assert not self.video is None, "When calling DetectionResult.check() specify the gt_path or use a video like v.detect().check()"

            assert self.video.has_meteors(), f"Video {self.video} has no gts in our data base. Specify a meteors file with the `gt_path` argument"

            return self.video.evaluate_args(self.args, self.video.meteors(), stdout=stdout, log=log)

        else:
            return fmdt.check(self.args.detect_args.trk_out_path, gt_path, stdout=stdout, log=log)

    # @staticmethod
    # def from_file(trk_path: str, log_path: str):

    
    # @staticmethod
    # def from_detect(args: fmdt.args.Args):

    #     trk_path = args.detect_args.trk_out_path
    #     log_path = args.detect_args.log_path

    #     if trk_path is None:
    #         trk_list = []
    #     else:
    #         trk_list = fmdt.extract_all_information(trk_path)

    #     if not log_path is None:
    #         rois, nassocs, mean_errs, std_devs = retrieve_log_info(log_path) 
    #     else:
    #         rois = []
    #         nassocs = []
    #         mean_errs = []
    #         std_devs = []

    #     return DetectionResult(rois, [0] + nassocs, [0.0] + mean_errs, [0.0] + std_devs, args, trk_list)
        
    
# load a DetectionResult object from a trk_path and a log_path
def load_det_result(trk_path: str, log_path: str) -> DetectionResult:
    """Load a det_result object whose content is stored in trk_path and log_path"""

    if os.path.exists(trk_path):
        trk_list = fmdt.core.extract_all_information(trk_path) 
        n_frames = fmdt.core.nframes_processed(trk_path)
    else:
        trk_list = []
        n_frames = 0

    if os.path.exists(log_path):
        df = retrieve_log_df(log_path)
    else:
        df = None

    return DetectionResult(n_frames, df, None, trk_list)

#===================== Fonctions dealing with check ============================

# I want to parse the file.
# Let's start off by parsing the first table

# start after Id|    Type 
# end at Statistics

CHECK_TABLE_HEADER = "#   Id |    Type || Detect |  GT || Start |  Stop ||      #"

CTBL_ID = 0
CTBL_TYPE = 2
CTBL_DETECT = 4
CTBL_GT = 6
CTBL_START = 8
CTBL_STOP = 10
CTBL_TRACKS = 12

def load_check_gt_table(check_stdout: str) -> pd.DataFrame:


    with open(check_stdout) as file:
        lines = file.readlines()

        hdr_line_nmbr  = -1
        stat_line_nmbr = -1

        # iterate through lines until the we get to the Id line
        for (i, el) in enumerate(lines):
            if el.strip() == CHECK_TABLE_HEADER:
                hdr_line_nmbr = i
                break

        for (i, el) in enumerate(lines):
            if "Statistics:" in el:
                stat_line_nmbr= i 

        table = lines[(hdr_line_nmbr + 2):stat_line_nmbr] 


        ids = []
        types = []
        detects = []
        gts = []
        starts = []
        stops = []
        tracks = []

        for line in table:
            s = line.split()

            ids.append(int(s[CTBL_ID]))
            types.append(s[CTBL_TYPE])
            detects.append(int(s[CTBL_DETECT]))
            gts.append(int(s[CTBL_GT]))
            starts.append(int(s[CTBL_START]))
            stops.append(int(s[CTBL_STOP]))
            tracks.append(int(s[CTBL_TRACKS]))
        
    return pd.DataFrame({
        "id": ids,
        "types": types,
        "detects": detects,
        "gts": gts,
        "starts": starts,
        "stops": stops,
        "tracks": tracks
    })

CHECK_NGT = "- Number of GT objs"      
CHECK_NTRACK = "- Number of tracks"
CHECK_TPOS = "- True positives"
CHECK_FPOS = "- False positives"
CHECK_TNEG = "- True negative"
CHECK_FNEG = "- False negative"
CHECK_TRACK_RATE = "- tracking rate"

def check_line_to_tuple(line: str) -> tuple[int, int, int, int]:
    """Convert '- Number of GT objs = ['meteor':   34, 'star':    0, 'noise':    0, 'all':   34]' to a 4-tuple of numbers """

    # Find the first instance of []
    line = line.strip()

    for (i, el) in enumerate(line):
        if line[i] == "[":
            break

    array_interior = line[i + 1:-1] # drop the brackets

    content = array_interior.split()

    m = int(content[1][:-1]) # drop comma at the end
    s = int(content[3][:-1]) # drop comma at the end
    n = int(content[5][:-1])
    a = int(content[7])

    return m, s, n, a

def check_tracking_rate_line(line: str) -> tuple[int, int, int, int]:
    """Convert '- Number of GT objs = ['meteor':   34, 'star':    0, 'noise':    0, 'all':   34]' to a 4-tuple of numbers """

    # Find the first instance of []
    line = line.strip()

    for (i, el) in enumerate(line):
        if line[i] == "[":
            break

    array_interior = line[i + 1:-1] # drop the brackets

    content = array_interior.split()

    m = float(content[1][:-1]) # drop comma at the end
    s = float(content[3][:-1]) # drop comma at the end
    n = float(content[5][:-1])
    a = float(content[7])

    return m, s, n, a


def load_check_stats(check_stdout: str) -> pd.DataFrame:

    with open(check_stdout) as file:
        lines = file.readlines()

        # iterate through lines until the we get to the Id line
        for (i, el) in enumerate(lines):
            if CHECK_NGT in el:
                break

        ms = []; ss = []; ns = []; alls = [];
        mat = [ms, ss, ns, alls] # these are our rows and we need to get the matrix as column

        for j in range(6):
            m, s, n, a = check_line_to_tuple(lines[i + j])
            ms.append(m); ss.append(s); ns.append(n); alls.append(a)
        
        m, s, n, a = check_tracking_rate_line(lines[i + 6])
        ms.append(m); ss.append(s); ns.append(n); alls.append(a)

        ngts = []; ntrks = []; tpos = []; fpos = []; tneg = []; fneg = []; trk_rate = []

        names = ["meteor", "star", "noise", "all"]

        for r in mat:

            ngts.append(r[0])
            ntrks.append(r[1])
            tpos.append(r[2])
            fpos.append(r[3])
            tneg.append(r[4])
            fneg.append(r[5])
            trk_rate.append(r[6])
    
    return pd.DataFrame({
        "type": names,
        "gt": ngts,
        "ntrk": ntrks,
        "tpos": tpos,
        "fpos": fpos,
        "tneg": tneg,
        "fneg": fneg,
        "trk_rate": trk_rate
    })

class CheckResult:

    def __init__(self, gt_table: pd.DataFrame = None, stats: pd.DataFrame = None):
        self.gt_table = gt_table
        self.stats = stats

    def meteor_stats(self) -> pd.Series:
        return self.stats[self.stats["type"] == "meteor"].iloc[0]

    def noise_stats(self) -> pd.Series:
        return self.stats[self.stats["type"] == "noise"].iloc[0]

    def star_stats(self) -> pd.Series:
        return self.stats[self.stats["type"] == "star"].iloc[0]

    def all_stats(self) -> pd.Series:
        return self.stats[self.stats["type"] == "all"].iloc[0]


def main():
    file = "test_check_one.txt"

    df = load_check_gt_table(file) 





"""Module dealing with the database stuff"""

import fmdt.config
import fmdt.args
import fmdt.truth
import fmdt.download
import fmdt.utils
import fmdt.res
import fmdt.api

import pandas as pd
import sqlite3
from enum import Enum
import os

VIDEOS_FILE = fmdt.config.dir() + "/videos.db"

class VideoType(Enum):
    DRACO6 = 0,
    DRACO12 = 1,
    WINDOW = 2,
    OTHER = 3

    def __str__(self) -> str:
        if self == VideoType.DRACO6:
            return "DRACO6"
        elif self == VideoType.DRACO12:
            return "DRACO12"
        elif self == VideoType.WINDOW:
            return "WINDOW"
        else:
            return "OTHER"
        
    def __repr__(self) -> str:
        return self.__str__()
    
    def dir(self) -> str:

        con = fmdt.config.load_config()
        if self == VideoType.DRACO6:
            return con.d6
        elif self == VideoType.DRACO12:
            return con.d12
        elif self == VideoType.WINDOW:
            return con.win
        else:
            return "./"
        
    
    @staticmethod
    def from_str(str):
        if str == "DRACO6":
            return VideoType.DRACO6 
        elif str == "DRACO12":
            return VideoType.DRACO12
        else:
            return VideoType.WINDOW

# Or maybe recording, night sky watching, stargazing, video of night sky
# type is an enum { draco6, draco12, window}
class Video:

    csv_hdr = "id,name,type\n"
    
    def __init__(self, name: str, type: VideoType = None):
        self.name = name
        self.type = type

    def to_csv(self) -> str:
        return f"{self.name},{self.type}\n"
    
    def __str__(self) -> str:
        return f"{self.name}"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def is_draco6(self) -> bool:
        return self.type == VideoType.DRACO6
    
    def is_window(self) -> bool:
        return self.type == VideoType.WINDOW
    
    def is_draco12(self) -> bool:
        return self.type == VideoType.DRACO12
    
    def dir(self) -> str:
        return self.type.dir()
    
    def visu_name(self) -> str:

        beg, ext = fmdt.utils.decompose_video_filename(self.name)

        return beg + "_visu." + ext
    
    def prefix(self) -> str:

        pre, _ = fmdt.utils.decompose_video_filename(self.name)
        return pre

    def suffix(self) -> str:
        _, suffix = fmdt.utils.decompose_video_filename(self.name)
        return suffix
    
    def default_trk_path(self, dir: str = "."):
        return dir + "/" + self.prefix() + "_trk.txt" 



    # def visu
    def detect(self,
        #=================== fmdt-detect parameters ================
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = "bb.txt",
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        #================== Additional Parameters ====================
        trk_out_path: str | None = "trk.txt",
        log: bool = False,
        timeout: float = None,
        #================== Parameters for logging ===================
        cache: bool = False,
        save_df: bool = False
    ) -> fmdt.res.DetectionResult:
        """Call fmdt-detect with the provided parameters"""

        args = fmdt.args.detect_args(self.full_path(), vid_in_start, vid_in_stop, 
        vid_in_skip, vid_in_buff, vid_in_loop, vid_in_threads, light_min, light_max, 
        ccl_fra_path, ccl_fra_id, mrp_s_min, mrp_s_max, knn_k, knn_d, knn_s, trk_ext_d,
        trk_ext_o, trk_angle, trk_star_min, trk_meteor_min, trk_meteor_max, trk_ddev, 
        trk_all, trk_bb_path, trk_mag_path, log_path, trk_out_path, log, timeout=timeout)

        res = args.detect(cache=cache, save_df=save_df)
        res.video = self

        return res

# import os

    # def check(
    #         self,
    #         **kwargs
    #         # args: fmdt.args.Args = None # Optional
    #     ) -> fmdt.res.CheckingResult:
    #     """Call fmdt-check after a call to fmdt-detect with detection parameters `args`"""

    #     assert self.has_meteors(), f"Cannot call Video.check() because {self.name} has no ground truths in our database"

    #     args = fmdt.detect_args(**kwargs)
        
    #     # I want to see if we have a cached trk list
    #     cached_trk_path = args.detect_args.cache_trk()
    #     cached_met_path = args.detect_args.cache_dir() + "_meteors.txt"


    #     if not os.path.exists(cached_trk_path):
    #         # Then we can directly call fmdt.check
    #         args.detect()
    #         # gen tmp file using meteors
        

    #     pass


    
    def check_args(self, args: fmdt.Args) -> list[bool]: 
        """Check which meteors are detected by this choice of args using our internal python implementation to check if
        a meteor and tracked object are the same

        Returns
        -------

        success_list (list[bool]): a list indicating which meteors were detected by this set of parameters
        """

        args.detect_args.vid_in_path = self.full_path()
        res = args.detect()

        return [m.is_detected_in_list(res.trk_list) for m in self.meteors()] 
    



    def evaluate(self,        
        #=================== fmdt-detect parameters ================
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = "bb.txt",
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        #================== Additional Parameters ====================
        trk_out_path: str | None = "trk.txt",
        log: bool = False,
        timeout: float = None,
        #================== Check ====================================
        stdout: str = None,
        log_check = False
        ):

        args = fmdt.args.detect_args(self.full_path(), vid_in_start, vid_in_stop, 
        vid_in_skip, vid_in_buff, vid_in_loop, vid_in_threads, light_min, light_max, 
        ccl_fra_path, ccl_fra_id, mrp_s_min, mrp_s_max, knn_k, knn_d, knn_s, trk_ext_d,
        trk_ext_o, trk_angle, trk_star_min, trk_meteor_min, trk_meteor_max, trk_ddev, 
        trk_all, trk_bb_path, trk_mag_path, log_path, trk_out_path, log, timeout=timeout)

        res = args.detect()

        self.evaluate_args(args, res.trk_list, stdout, log=log_check)



    # Lookup the id in our default database file.
    def id(self) -> int:

        con = sqlite3.connect(VIDEOS_FILE)

        query = f"""
            select id from video where name = '{self.name}' 
        
        """

        df = pd.read_sql_query(query, con)
        con.close()        

        assert len(df) == 1, f"Something went wrong when looking up {self.name}"
        # Alternatively, we could return -1...
        
        return df.iat[0,0]
    
    def full_path(self) -> str:
        return self.dir() + "/" + self.name
    
    def meteors(self) -> list[fmdt.HumanDetection]:
        return retrieve_meteors(self.name)
    
    def has_meteors(self) -> bool:
        return len(self.meteors()) > 0

    def exists(self) -> bool:
        """Check whether the video path in self.full_path() exists on disk"""
        return os.path.exists(self.full_path())

    def evaluate_args(
            self,
            args: fmdt.Args,
            meteors: list[fmdt.truth.HumanDetection],
            rerun: bool = False,
            tmp_gt_file = "tmp_meteors.txt",
            stdout: str = None,
            log = False
        ):
        """Call fmdt-detect and fmdt-check to evaluate how well a given set of arguments detects our ground truth
        
        Parameters
        ----------
        
        args (fmdt.Args): The set of fmdt-detect parameters that we want to evaluate
        meteors (list[fmdt.HumanDetection]): The list of meteors in our ground truth
        rerun (bool): Used to determine if we should rerun fmdt-detect when the trk_out_path file
            already exists
        """
        assert not args.detect_args.trk_out_path is None, "Missing `trk_out_path` from `args: fmdt.Args` required for evaluation"
        # assert not args.detect_args.vid_in_path  is None, "Missing `vid_in_path` required to evaluate an fmdt.Args"
        args.detect_args.vid_in_path = self.full_path()

        if rerun:
            args.detect()

        # Check if the trk_out_path file already exists
        if not os.path.exists(args.detect_args.trk_out_path):
            args.detect()

        # We guarentee that the file exist at this point

        # Let's write the data from meteors to a temp file
        fmdt.truth.save_meteors_file(tmp_gt_file, meteors)

        # Then call fmdt_check

        return fmdt.check(args.detect_args.trk_out_path, tmp_gt_file, stdout, log)


    @staticmethod
    def from_pd_row(ser: pd.Series):
        out = Video(ser["name"], VideoType.from_str(ser["type"]))
        return out
    


# Take a list of Videos and turn it into a data base.
def videos_to_csv(videos: list[Video], csv_filename: str) -> None:
    # First get the header
    hdr = "id,name,type\n"

    with open(csv_filename, "w") as csv:
        csv.write(hdr)
        
        i = 0
        for v in videos:
            csv.write(f"{i},{v.name},{v.type}\n")
            i += 1

def csv_to_videos(csv_filename: str) -> list[Video]:

    df = pd.read_csv(csv_filename)
    return [Video.from_pd_row(df.iloc[i]) for i in range(len(df))]


def load_in_videos(db_filename: str = "videos.db", dir = fmdt.download.__DATA_DIR) -> list[Video]:
    """Read in the videos stored in 'videos.db' into a list of fmdt.Video"""

    # Download if the database file requested doesnt exist
    if not os.path.exists(dir + "/" + db_filename):
        fmdt.download.download_videos_db(db_filename, log=False, overwrite=False, dir=dir)

    if not dir is None:
        db_filename = dir + "/" + db_filename

    con = sqlite3.connect(db_filename)
    df = pd.read_sql_query("select * from video", con)
    con.close()

    return [fmdt.Video.from_pd_row(df.iloc[i]) for i in range(len(df))]

def has_meteors(vids: list[Video]) -> list[Video]:
    return [v for v in vids if v.has_meteors()]

def load_draco6(filename: str = "videos.db", db_dir = fmdt.download.__DATA_DIR, require_gt = False) -> list[Video]:
    """Load draco6 `Video` objects that are stored in the `db_dir`/`filename` .db file"""
    vids = load_in_videos(filename, db_dir)
    d6 = [v for v in vids if v.is_draco6()]
    if require_gt:
        return [v for v in d6 if v.has_meteors()]
    else:
        return d6

def load_draco12(db_filename: str = "videos.db", db_dir = fmdt.download.__DATA_DIR, require_gt = False) -> list[Video]:
    vids = load_in_videos(db_filename, db_dir)
    d12 = [v for v in vids if v.is_draco12()]
    if require_gt:
        return [v for v in d12 if v.has_meteors()]
    else:
        return d12

def load_window(db_filename: str = "videos.db", db_dir = fmdt.download.__DATA_DIR, require_gt = False) -> list[Video]:
    vids = load_in_videos(db_filename, db_dir)
    win = [v for v in vids if v.is_window()]
    if require_gt:
        return [v for v in win if v.has_meteors()]
    else:
        return win
    
def load_demo(db_filename: str = "videos.db", db_dir = fmdt.download.__DATA_DIR) -> list[Video]:
    vids = load_in_videos(db_filename, db_dir)
    win = [v for v in vids if v.name == "2022_05_31_tauh_34_meteors.mp4"]

    return win[0]


def retrieve_meteors(video_name: str, db_filename: str = "videos.db", dir = fmdt.download.__DATA_DIR) -> list[fmdt.HumanDetection]:

    if not os.path.exists(dir + "/" + db_filename):
        fmdt.download.download_videos_db(db_filename, log=False, overwrite=False, dir=dir)

    if not dir is None:
        db_filename = dir + "/" + db_filename

    query = f"""
        select * from human_detection where video_name = '{video_name}'
    """

    con = sqlite3.connect(db_filename)
    df = pd.read_sql_query(query, con)
    con.close()

    return [fmdt.HumanDetection.from_pd_row(df.iloc[i]) for i in range(len(df))]
    # return df

def get_video_diagnostics(vids: list[Video]) -> tuple[int, int]:
    """Print information about the local environment"""
    # d6 = fmdt.load_draco6()/
    on_disc = [v for v in vids if v.exists()]
    has_meteors = [v for v in on_disc if v.has_meteors()]

    n_exist = len(on_disc)
    n_with_meteors = len(has_meteors)

    return n_exist, n_with_meteors

def print_diagnostics(vids: list[Video]):

    n_exist, n_with_meteors = get_video_diagnostics(vids)


    print(f"{n_exist} videos exist on disc out of the {len(vids)} videos in our database")
    print(f"{n_with_meteors} of which have ground truths out of {len([v for v in vids if v.has_meteors()])} ground truths in our database")

def print_diagnostics_d6():
    print("================================================================================")
    print(f"Draconids-6mm*.avi videos configured with dir: {fmdt.config.draco6_dir()}")
    print("================================================================================")
    print_diagnostics(fmdt.load_draco6())
    
def print_diagnostics_d12():
    print("================================================================================")
    print(f"Draconids-12mm*.avi videos configured with dir: {fmdt.config.draco12_dir()}")
    print("================================================================================")
    print_diagnostics(fmdt.load_draco12())

def print_diagnostics_win():
    print("================================================================================")
    print(f"window*.mp4 videos configured with dir: {fmdt.config.window_dir()}")
    print("================================================================================")
    print_diagnostics(fmdt.load_window())

def print_diagnostics_all():
    """Print diagnostic information for the three class of videos in our database, Draco6, Draco12, and window

    We compare the number of videos that are present on disk with the number of videos in our database. 
    
    """
    print("Printing information about the local environment")
    print_diagnostics_d6()
    print()
    print_diagnostics_d12()
    print()
    print_diagnostics_win()
    print()

def info():
    print_diagnostics_all()

#!============================== Modifying our data base ============================
def add_human_detection_to_gt(meteor: fmdt.HumanDetection):
    pass
    


"""This script shows you how to load in our GroundTruth databases"""
import fmdt

# Specify where your videos are stored
fmdt.init(d6_dir = "/home/ejovo/Videos/Watec6mm",
          d12_dir = "/home/ejovo/Videos/Watec12mm",
          win_dir = "/home/ejovo/Videos/Window")

# Load in the database
gt6 = fmdt.load_gt6()
gt12 = fmdt.load_gt12()

# Far cleaner than before!
args = fmdt.detect_args(light_min=254, light_max=255, timeout=0.5)

# gt6.try_command(args)

# success_list = gt6.try_command(args)

# print(success_list)
# print(sum(success_list))

gt6.draw_heatmap(150, 200, 2)


"""
API to call fmdt executables. Assumes that fmdt-detect and other executables
are on the system path
"""
import fmdt.res
import random
from os import listdir
import subprocess
import os
import fmdt.core
import fmdt.utils
import fmdt.args
from termcolor import colored
import shutil
import pandas as pd

def help():
    s = """    fmdt.api contains a Python wrapper for the `fmdt-detect` and `fmdt-visu` 
    executables.
    
    fmdt.api contains the public-facing functions (also aliased under fmdt.*):
        fmdt[.api].count
        fmdt[.api].detect
        fmdt[.api].visu
        fmdt[.api].detect_directory

    fmdt.api.count counts the number of celestial objects specified by the parameters
    fmdt.api.detect calls `fmdt-detect` with the given arguments
    fmdt.api.visu calls `fmdt-visu`
    fmdt.api.detect_directory
        
        
        
"""

def count(
        trk_path: str,
        stars: bool = False,
        meteors: bool = True,
        noise: bool = False,
        all: bool = False
    ) -> int:
    """Count the number of objects detected in a tracks_bb file

    Parameters
    ----------
    trk_path (str): Path to file that stores the tracks (stdout of `fmdt-detect`) of detected objects
    stars (bool): When True, include star objects in our count (Default False)
    meteors (bool): When True, include meteor objects in our count (Default True)
    noise (bool): When True, include noise objects in our count (Default False)
    all (bool): When True, include all objects in the count
    
    Examples
    --------
    >>> fmdt.detect(vid_in_path="demo.mp4", trk_out_path="tracks.txt") # Store tracks in tracking file
    >>> fmdt.count("tracks.txt")  # Count meteors only

    >>> fmdt.count("tracks.txt", all=True) # Count stars, meteors, and noise

    >>> fmdt.count("tracks.txt", stars=True, meteors=False, noise=True) # Count stars and noise 
    """
    if all:
        stars = True
        meteors = True
        noise = True

    tracked = fmdt.core.extract_key_information(trk_path)

    # If we shouldn't keep this object type, filter it out
    def refine(obj_type: bool, type_str: str) -> None: 
        if not obj_type:
            tracked = [t for t in tracked if t["type"] != type_str]
    
    refine(stars, "stars")
    refine(meteors, "meteors")
    refine(noise, "noise")

    # tracked_meteors = fmdt.utils.retain_meteors(tracked)
    return len(tracked)
    

FMDT_TIMEOUT = 1

def _run_detect_stdout(
        argv: list[str],
        timeout: float,
        log: bool,
        cache: bool,
        cache_file: str = None
    ) -> tuple[list[fmdt.core.TrackedObject], int]:
    """Handle the final logic of calling `fmdt-detect` when not storing the track list in a file"""

    if log:
        print(f"Executing cmd: {' '.join(argv)}")

    # Store this data in a tmp file and then use that to get the tracking list.
    tmp_filename = "fmdt_detect_tmp.txt"
    tmp_file = open(tmp_filename, "w")

    if not timeout is None:

        try:
            subprocess.run(argv, timeout=timeout, stdout=tmp_file)
            tmp_file.close()

        except:
            print(f"Subprocess timed out for {colored(' '.join(argv), 'red')}")
            tmp_file.close()
            os.remove(tmp_file)
            return [], 0
    else:
        subprocess.run(argv, stdout=tmp_file)

    if log: 
        with open(tmp_filename) as tmp_file:
            print(tmp_file.read())
    
    if cache:
        shutil.copyfile(src=tmp_file, dst=cache_file)

    trk_list = fmdt.core.extract_all_information(tmp_filename)
    nframes = fmdt.core.nframes_processed(tmp_filename)

    os.remove(tmp_filename)

    return trk_list, nframes

def _run_detect_trk_path(
        trk_out_path: str,
        argv: list[str],
        timeout: float,
        log: bool,
        cache: bool,
        cache_file: str
    ) -> tuple[list[fmdt.core.TrackedObject], int]:

    with open(trk_out_path, 'w') as outfile:

        if log:
            print(f"Executing cmd: {' '.join(argv)}")

        if not timeout is None:
            try: 
                subprocess.run(argv, stdout=outfile, timeout=timeout)
            except:
                print("==================================================================")
                print("")
                print(f"Subprocess timed out for \n\t{colored(' '.join(argv), 'blue')}")
                print("")
                print("==================================================================")
                return [], 0
        else:
            subprocess.run(argv, stdout=outfile)
        
        trk_list = fmdt.core.extract_all_information(trk_out_path)
        nframes = fmdt.core.nframes_processed(trk_out_path)
            
    if cache:
        shutil.copyfile(src=trk_out_path, dst=cache_file)

    return trk_list, nframes
    

def detect(
        #=================== fmdt-detect parameters ================
        vid_in_path: str, 
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = "bb.txt",
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        #================== Additional Parameters ====================
        trk_out_path: str | None = "trk.txt",
        log: bool = False,
        timeout: float = None,
        cache: bool = False,
        save_df: bool = False
    ) -> fmdt.res.DetectionResult:
    """Wrapper to executable fmdt-detect.

    Parameters
    ----------
    Extensively documented here: https://fmdt.readthedocs.io/en/latest/user/usage/detect.html#

    Wrapper Parameters
    ------------------
    trk_out_path (str): path to redirect stdout of `fmdt-detect`
    log (bool): Print logging messages to stdout (True) or do nothing (False). Default False.
    timeout (float): timeout in seconds of the Python subprocess executing `fmdt-detect`. Default None. 
        Used to speed up ground truth testing.
    """


    # Wrap up all of the arguments into an Args object
    args = fmdt.args.detect_args(vid_in_path, vid_in_start, vid_in_stop, 
        vid_in_skip, vid_in_buff, vid_in_loop, vid_in_threads, light_min, light_max, 
        ccl_fra_path, ccl_fra_id, mrp_s_min, mrp_s_max, knn_k, knn_d, knn_s, trk_ext_d,
        trk_ext_o, trk_angle, trk_star_min, trk_meteor_min, trk_meteor_max, trk_ddev, 
        trk_all, trk_bb_path, trk_mag_path, log_path, trk_out_path, log)
    
    if save_df and log_path is None:
        print("Save_df activated in final detect call")
        args.detect_args.log_path = args.detect_args.cache_dir()

    # Spit out the commandline arguments for fmdt-detect
    argv = args.detect_args.argv()
    cache_file = None

    if cache:
        cache_file = args.detect_args.cache_trk()

    #============ Retrieve Tracked list ===========================================#
    if trk_out_path is None:
        trk_list, nframes = _run_detect_stdout(argv, timeout, log, cache, cache_file) 
    else:
        trk_list, nframes = _run_detect_trk_path(trk_out_path, argv, timeout, log, cache, cache_file) 

    #============= Recover data if log_path =======================================#
    if not args.detect_args.log_path is None:
        nrois, nassocs, mean_errs, std_devs = fmdt.res.retrieve_log_info(args.detect_args.log_path) 
    else:
        nrois = []
        nassocs = []
        mean_errs = []
        std_devs = []

    # Now construct the result object

    if len(nrois) == 0:
        df = None
    else:
        df =  pd.DataFrame({
                'nroi': nrois,
                'nassoc': [0] + nassocs,
                'mean_err': [0.0] + mean_errs,
                'std_dev': [0.0] + std_devs
            })

    return fmdt.res.DetectionResult(nframes, df, args, trk_list)   
            
def visu(
        vid_in_path: str = None,
        vid_in_start: int = None,
        vid_in_stop: int = None,
        vid_in_threads: int = None,
        trk_path: str = None,
        trk_bb_path: str = None,
        trk_id: bool = None,
        trk_nat_num: bool = None,
        trk_only_meteor: bool = None,
        gt_path: str = None,
        vid_out_path: str = None
    ) -> fmdt.args.Args:
    """Wrapper to executable fmdt-detect.

    Parameters
    ----------
    Extensively documented here: https://fmdt.readthedocs.io/en/latest/user/usage/visu.html
    """
    visu_args = fmdt.args.visu_args(vid_in_path, 
                                    vid_in_start,
                                    vid_in_stop,
                                    vid_in_threads,
                                    trk_path,
                                    trk_bb_path,
                                    trk_id,
                                    trk_nat_num,
                                    trk_only_meteor,
                                    gt_path,
                                    vid_out_path)

    argv = fmdt.args.handle_visu_args(**visu_args.to_dict())

    subprocess.run(argv)

    return fmdt.args.Args(visu_args=visu_args, detect_args=None)

def check(
    trk_path: str, gt_path: str, stdout: str = None, log = False
    ) -> fmdt.res.CheckResult:
    """Call fmdt-check
    
    Parameters
    ----------
    std_out (str): File to store stdout of fmdt-check"""

    argv = ["fmdt-check", "--trk-path", trk_path, "--gt-path", gt_path]
    res = subprocess.run(argv, stdout=subprocess.PIPE)

    file_content = res.stdout.decode()

    if not stdout is None:
        with open(stdout, "w") as file:
            file.write(file_content)

        stats = fmdt.res.load_check_stats(stdout)
        gt_table = fmdt.res.load_check_gt_table(stdout)
    
    else:

        stats = None
        gt_table = None

    if (log):
        print(file_content)

    return fmdt.res.CheckResult(gt_table=gt_table, stats=stats)



def detect_directory(dir_name: str, args: fmdt.args.Args, log=False):
    """Call `fmdt-detect` on all videos in the directory `dir_name` using the settings stored in `args`
    
    Parameters
    ----------
    dir_name (str): Path to the directory of videos that you'd like to detect
    args: (fmdt.args.Args): Configuration of parameters used to call fmdt.detect    
    
    
    """

    entries = listdir(dir_name)
    is_video_fn = lambda v: v[-3:] == "mp4" or v[-3:] == "avi"
    videos = [e for e in entries if is_video_fn(e)]

    assert len(videos) > 0, "Directory is empty, call to fmdt.detect_directory failed"

    # Now let's call fmdt detect one time for each video
    failing_cmds = []
    i = 0
    for v in videos:
        # res = fmdt.detect(directory + "/" + v, light_min=150, light_max=245, trk_all=True, log=False, trk_out_path="tracks.txt")
        # a = fmdt.args.video_input(directory + "/" + v)
        # a.detect_args["trk_out_path"] = "tracks.txt"

        # if i > 10:
        #     break
        args.detect_args["vid_in_path"] = dir_name + "/" + v

        fail = args.does_detect_fail(log=log)
        if log: 
            print(f"{v} fails? {fail}")

        if (fail):
            failing_cmds.append(" ".join(args.detect_cmd()))
        
        i = i + 1

    for c in failing_cmds:
        print(c)


import os
import fmdt
import fmdt.args
import sqlite3

w6 = "/run/media/ejovo/Seagate Portable Drive/Meteors/Watec6mm/Meteor"
w12 = "/run/media/ejovo/Seagate Portable Drive/Meteors/Watec12mm/Meteor"
win = "/run/media/ejovo/Seagate Portable Drive/Meteors"


def is_video(vid_name: str) -> bool:
    return vid_name[-3:] == "avi" or vid_name[-3:] == "mp4"

def is_draco6(vid_name: str) -> bool:
    return is_video(vid_name) and vid_name[0:13] == "Draconids-6mm"

def is_draco12(vid_name: str) -> bool:
    return is_video(vid_name) and vid_name[0:14] == "Draconids-12mm"

def is_window(vid_name: str) -> bool:
    return is_video(vid_name) and vid_name[0:6] == "window"


# def get_vids(dir: str, type: fmdt.args.VideoType) -> list[]

def get_draco6(dir) -> list[fmdt.args.Video]:
    return [fmdt.Video(v, type=fmdt.args.VideoType.DRACO6) for v in os.listdir(dir) if is_draco6(v)]

def get_draco12(dir) -> list[fmdt.args.Video]:
    return [fmdt.Video(v, type=fmdt.args.VideoType.DRACO12) for v in os.listdir(dir) if is_draco12(v)]

def get_window(dir) -> list[fmdt.args.Video]:
    return [fmdt.Video(v, type=fmdt.args.VideoType.WINDOW) for v in os.listdir(dir) if is_window(v)]

d6 = get_draco6(w6)
d12 = get_draco12(w12)
dw = get_window(win)

d6.sort()
d12.sort()
dw.sort()

print(d6)
print()
print(d12)

print(len(d6))
print(len(d12))
print(len(dw))

all = [v for v in d6]
all.extend(d12)
all.extend(dw)

print(len(all))
print(type(all[0]))

fmdt.args.videos_to_csv(all, "test_vid.csv")
# videos = fmdt.args.csv_to_videos("test_vid.csv")
# print(len(videos))

# for i in range(10):





"""Play with the contents of `videos.db` programmatically and store results in data frames."""

import pandas as pd
import sqlite3
import fmdt.args
import fmdt.truth
import fmdt.config
import fmdt.db


# Add a videos.db file to our config
fmdt.download_csvs()

# config = fmdt.config.load_config()
# files_in_config = fmdt.config.listdir()

# print(files_in_config)
print(f"config directory: {fmdt.config.dir()}")

fmdt.download_dbs()
vids = fmdt.load_draco6() # list[Video]
for i in range(20):
    v = vids[i]
    print(f"{v.id()}: {v} has {len(v.meteors())} meteor(s) in our database")

print(".")
print(".")
print(".")

print(f"{vids[-1].id()}: {vids[-1]} has {len(vids[-1].meteors())} meteor(s) in our database")

id = 17
v = vids[id]
print(f"vids[{id}].meteors() -> {v.meteors()[0]} and has full dir: {v.full_path()}")

on_disc = [v for v in vids if v.exists()]
has_meteors = [v for v in on_disc if v.has_meteors()]

print(f"{len(on_disc)} videos exist on disc out of {len(vids)}")
print(f"{len(has_meteors)} videos have meteors and exist on disk")

# Let's get some diagnostics then.

# Like: How many videos exist from our database exist on disk?

d6 = fmdt.load_draco6()
d12 = fmdt.load_draco12()
w = fmdt.load_window()


fmdt.db.info()

print(w)
print(w[0].full_path())


from fmdt.core import (
    extract_all_information,
    extract_key_information,
    split_video_at_meteors,
    split_video_at_intervals,
) 

from fmdt.api import (
    detect,
    visu,
    count,
    detect_directory,
    check
)

from fmdt.args import (
    Args,
    DetectArgs,
    VisuArgs,
    detect_args
)

from fmdt.utils import (
    condense_start_end
)

from fmdt.truth import (
    WATEC6_DIR,
    WATEC12_DIR,
    HumanDetection,
    read_human_detection_csv,
    init_ground_truth,
    GroundTruth
)

from fmdt.download import (
    download_csvs,
    download_demo_mp4,
    download_dbs
)

from fmdt.config import (
    init,
    load_gt12,
    load_gt6,
    load_config,
    listdir_draco12,
    listdir_draco6,
    listdir_window,
    setdir_draco6,
    setdir_draco12,
    setdir_window,
    init_cache,
    clear_cache,
    cache_dir,
    cache_info,
    listdir_cache,
    size_cache
)

from fmdt.db import (
    Video,
    VideoType,
    load_in_videos,
    load_draco6,
    load_draco12,
    load_window,
    load_demo,
    info as local_info
)

init_cache()



# from fmdt.truth import HumanDetection

MAJOR_VERSION = 0
MINOR_VERSION = 0
PATCH         = 33 

VERSION = str(MAJOR_VERSION) + "." + str(MINOR_VERSION) + "." + str(PATCH)



import fmdt
import os


def _mkdirp(dir: str):
    if not os.path.exists(dir):
        os.mkdir(dir)

# ============== Videos ==================================
d6 = fmdt.load_draco6(require_gt=True)
d12 = fmdt.load_draco12(require_gt=True)

# ============= GroundTruth (list[HumanDetection]) ======
gt6 = fmdt.load_gt6()
gt12 = fmdt.load_gt12()

print(d6)
print(d12)

print(gt6)
print(gt12)

args = fmdt.Args.new(light_min=253, light_max=255, timeout=0.5, log_path="log", log=True)

results = []

print(len(d6))

for i in range(len(d6)):
    v = d6[i]
    
    args.detect_args.vid_in_path = v.full_path()
    # args.detect_args.trk_out_path = v.default_trk_path()
    args.detect_args.trk_out_path = args.detect_args.cache_trk()
    args.detect_args.log_path = args.detect_args.cache_dir()

    _mkdirp(args.detect_args.cache_dir())
    results.append(args.detect())

for r in results:
    print(r.to_dataframe())

# list_bools = [v.check_args(args) for v in d6]

# for i in range(len(list_bools)):

#     def one_or_more_det(bools: list[bool]):
#         for b in bools:
#             if b:
#                 return b
    

# Keep the videos that had a detection with args

# vids_with_det = [d6[i] for i in range(len(list_bools)) if one_or_more_det(list_bools[i])]

# print(vids_with_det)
# print(f"# of vids with detected meteors: {len(vids_with_det)}")

# # I want to visualize each video that was detected

# for v in vids_with_det:
#     fmdt.visu(vid_in_path=v.full_path(), vid_out_path=v.visu_name(), )